EXTRAS_DOCS = "https://py.sdmx.io/start.html#how-can-i-get-it"
ERROR_MESSAGE = (
    "The '{extra_name}' extra is required to run {extra_desc}. "
    "Please install it using 'pip install pysdmx[{extra_name}]' or "
    "install all extras with 'pip install pysdmx[all]'. "
    f"Check the documentation at: {EXTRAS_DOCS}"
)


def __check_dc_extra() -> None:
    try:
        import dateutil  # noqa: F401
    except ImportError:
        raise ImportError(
            ERROR_MESSAGE.format(
                extra_name="dc", extra_desc="the python parser for dates"
            )
        ) from None


def __check_data_extra() -> None:
    try:
        import pandas  # noqa: F401
    except ImportError:
        raise ImportError(
            ERROR_MESSAGE.format(
                extra_name="data",
                extra_desc="the reading and writing of Data Messages",
            )
        ) from None


def __check_xml_extra() -> None:
    try:
        import lxml  # noqa: F401
        import sdmxschemas  # noqa: F401
        import xmltodict  # noqa: F401
    except ImportError:
        raise ImportError(
            ERROR_MESSAGE.format(
                extra_name="xml",
                extra_desc="the reading and writing of SDMX-ML Messages",
            )
        ) from None


def __check_vtl_extra() -> None:
    try:
        import vtlengine  # type: ignore[import-untyped]  # noqa: F401
    except ImportError:
        raise ImportError(
            ERROR_MESSAGE.format(
                extra_name="vtlengine",
                extra_desc="validations for VTL engine",
            )
        ) from None



"""Collection of Fusion-JSON schemas for organisations."""

from collections import defaultdict
from typing import Dict, Optional, Sequence, Set

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.io.json.fusion.messages.pa import FusionProvisionAgreement
from pysdmx.model import (
    Agency,
    Contact,
    DataflowRef,
    DataProvider,
)
from pysdmx.model import (
    AgencyScheme as AS,
)
from pysdmx.model import (
    DataProviderScheme as DPS,
)
from pysdmx.util import parse_item_urn, parse_urn


class FusionContact(Struct, frozen=True):
    """Fusion-JSON payload for a contact."""

    id: Optional[str] = None
    names: Optional[Sequence[FusionString]] = None
    departments: Optional[Sequence[FusionString]] = None
    roles: Optional[Sequence[FusionString]] = None
    telephone: Optional[Sequence[str]] = None
    fax: Optional[Sequence[str]] = None
    uri: Optional[Sequence[str]] = None
    email: Optional[Sequence[str]] = None

    def to_model(self) -> Contact:
        """Converts a FusionContact to a standard Contact."""
        n = self.names[0].value if self.names else None
        d = self.departments[0].value if self.departments else None
        r = self.roles[0].value if self.roles else None

        return Contact(
            self.id,
            n,
            d,
            r,
            self.telephone,
            self.fax,
            self.uri,
            self.email,
        )


class FusionAgency(Struct, frozen=True):
    """Fusion-JSON payload for an organisation."""

    id: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    contacts: Sequence[FusionContact] = ()

    def to_model(self, owner: Optional[str] = None) -> Agency:
        """Converts a FusionOrg to a standard Organisation."""
        d = self.descriptions[0].value if self.descriptions else None
        c = [c.to_model() for c in self.contacts]
        oid = f"{owner}.{self.id}" if owner and owner != "SDMX" else self.id
        if c:
            return Agency(
                id=oid, name=self.names[0].value, description=d, contacts=c
            )
        else:
            return Agency(id=oid, name=self.names[0].value, description=d)


class FusionProvider(Struct, frozen=True):
    """Fusion-JSON payload for an organisation."""

    id: str
    names: Sequence[FusionString]
    descriptions: Optional[Sequence[FusionString]] = None
    contacts: Sequence[FusionContact] = ()

    def to_model(self, owner: Optional[str] = None) -> DataProvider:
        """Converts a FusionOrg to a standard Organisation."""
        d = self.descriptions[0].value if self.descriptions else None
        c = [c.to_model() for c in self.contacts]
        oid = f"{owner}.{self.id}" if owner and owner != "SDMX" else self.id
        if c:
            return DataProvider(
                id=oid, name=self.names[0].value, description=d, contacts=c
            )
        else:
            return DataProvider(
                id=oid, name=self.names[0].value, description=d
            )


class FusionAgencyScheme(Struct, frozen=True):
    """Fusion-JSON payload for an agency scheme."""

    agencyId: str
    descriptions: Sequence[FusionString] = ()
    items: Sequence[FusionAgency] = ()

    def to_model(self) -> AS:
        """Converts a FusionAgencyScheme to a list of Organisations."""
        agencies = [o.to_model(self.agencyId) for o in self.items]
        return AS(
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            agency=self.agencyId,
            items=agencies,
        )


class FusionAgencyMessage(Struct, frozen=True):
    """Fusion-JSON payload for /agencyscheme queries."""

    AgencyScheme: Sequence[FusionAgencyScheme]

    def to_model(self) -> Sequence[AS]:
        """Returns the requested agency schemes."""
        return [a.to_model() for a in self.AgencyScheme]


class FusionProviderScheme(Struct, frozen=True):
    """Fusion-JSON payload for a data provider scheme."""

    id: str
    agencyId: str
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()
    items: Sequence[FusionProvider] = ()

    def __get_df_ref(self, ref: str) -> DataflowRef:
        a = parse_urn(ref)
        return DataflowRef(id=a.id, agency=a.agency, version=a.version)

    def to_model(self, pas: Sequence[FusionProvisionAgreement]) -> DPS:
        """Converts a FusionProviderScheme to a DataProviderScheme."""
        if pas:
            paprs: Dict[str, Set[DataflowRef]] = defaultdict(set)
            for pa in pas:
                df = self.__get_df_ref(pa.structureUsage)
                ref = parse_item_urn(pa.dataproviderRef)
                paprs[f"{ref.agency}:{ref.item_id}"].add(df)
            prvs = [o.to_model() for o in self.items]
            prvs = [
                DataProvider(
                    id=p.id,
                    name=p.name,
                    description=p.description,
                    contacts=p.contacts,
                    dataflows=list(paprs[f"{self.agencyId}:{p.id}"]),
                )
                for p in prvs
            ]
            return DPS(
                agency=self.agencyId,
                name=self.names[0].value,
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                items=prvs,
            )
        else:
            return DPS(
                agency=self.agencyId,
                name=self.names[0].value,
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                items=[o.to_model() for o in self.items],
            )


class FusionProviderMessage(Struct, frozen=True):
    """Fusion-JSON payload for /dataproviderscheme queries."""

    DataProviderScheme: Sequence[FusionProviderScheme]
    ProvisionAgreement: Sequence[FusionProvisionAgreement] = ()

    def to_model(self) -> Sequence[DPS]:
        """Returns the requested list of providers."""
        return [
            p.to_model(self.ProvisionAgreement)
            for p in self.DataProviderScheme
        ]



"""SDMX 2.0 CSV reader and writer."""

from pysdmx.model.dataset import ActionType

SDMX_CSV_ACTION_MAPPER = {
    ActionType.Append: "A",
    ActionType.Replace: "R",
    ActionType.Information: "I",
    ActionType.Delete: "D",
}



from datetime import datetime
from enum import Enum
from typing import Sequence, Union

from msgspec import Struct

from pysdmx.api.dc.query._model import Operator


class _CoreOperator(Enum):
    EQUALS = "="
    NOT_EQUALS = "<>"
    LESS_THAN = "<"
    GREATER_THAN = ">"
    LESS_THAN_OR_EQUAL = "<="
    GREATER_THAN_OR_EQUAL = ">="
    LIKE = "LIKE"
    NOT_LIKE = "NOT LIKE"


class _InOperator(Enum):
    IN = "IN"
    NOT_IN = "NOT IN"


class _RangeOperator(Enum):
    BETWEEN = "BETWEEN"
    NOT_BETWEEN = "NOT BETWEEN"


class _NullOperator(Enum):
    NULL = "IS NULL"
    NOT_NULL = "IS NOT NULL"


class _Field(Struct, frozen=True):
    name: str


class _Number(Struct, frozen=True):
    value: Union[int, float]


class _String(Struct, frozen=True):
    value: str


class _Boolean(Struct, frozen=True):
    value: bool


class _DateTime(Struct, frozen=True):
    value: datetime


class _Filter(Struct, frozen=True):
    """A data filter."""

    field: _Field
    operator: Operator
    value: Union[
        None,
        _Boolean,
        _DateTime,
        _Number,
        Sequence[_DateTime],
        Sequence[_Number],
        Sequence[_String],
        _String,
    ] = None



"""Parsing configuration for SDMX-ML 2.1 messages."""

# Common
HEADER = "Header"
DATASET = "DataSet"
SERIES = "Series"
OBS = "Obs"
AGENCY_ID = "agencyID"
ID = "id"
VERSION = "version"
REQUIRED = "required"
POSITION = "position"
CLASS = "class"
PACKAGE = "package"

# Structure Specific
STR_SPE = "StructureSpecificData"
GENERIC = "GenericData"
SERIES_KEY = "SeriesKey"
ATTRIBUTES = "Attributes"
VALUE = "Value"
OBS_DIM = "ObsDimension"
OBS_VALUE_XML_TAG = "ObsValue"
OBS_VALUE_ID = "OBS_VALUE"
OBS_KEY = "ObsKey"
# Header
STR_REF = "structureRef"
STRUCTURE = "Structure"
STR_USAGE = "StructureUsage"
PROV_AGREMENT = "ProvisionAgrement"
STR_ID = "structureID"
STR_TYPE = "structure_type"
DIM_OBS = "dimensionAtObservation"
ALL_DIM = "AllDimensions"
REF = "Ref"
DATASET_ID = "DataSetID"
DATASET_ACTION = "DataSetAction"
TEST = "Test"
PREPARED = "Prepared"
SENDER = "Sender"
RECEIVER = "Receiver"
SOURCE = "Source"
HEADER_ID = "ID"
NAMESPACE = "namespace"
NAMES = "names"


# SDMX Error handling
ERROR = "Error"
ERROR_MESSAGE = "ErrorMessage"
ERROR_CODE = "code"
ERROR_TEXT = "Text"

# SDMX Registry Interface handling
REG_INTERFACE = "RegistryInterface"
SUBMIT_STRUCTURE_RESPONSE = "SubmitStructureResponse"
SUBMISSION_RESULT = "SubmissionResult"
SUBMITTED_STRUCTURE = "SubmittedStructure"
MAINTAINABLE_OBJECT = "MaintainableObject"
ACTION = "action"
STATUS_MSG = "StatusMessage"
STATUS = "status"

# Structures
# Common
NAME = "Name"
NAME_LOW = "name"
DESC = "Description"
DESC_LOW = "description"
LANG = "lang"
XML_TEXT = "#text"
STR_URL = "structureURL"
STR_URL_LOW = "structure_url"
SER_URL = "serviceURL"
SER_URL_LOW = "service_url"

# General
ANNOTATIONS = "Annotations"
ANNOTATIONS_LOW = "annotations"
STRUCTURES = "Structures"
ORGS = "OrganisationSchemes"
AGENCIES = "AgencyScheme"
CODELISTS = "Codelists"
CONCEPTS = "Concepts"
DSDS = "DataStructures"
DATAFLOWS = "Dataflows"
CONSTRAINTS = "Constraints"

# Individual
AGENCY = "Agency"
CL = "Codelist"
CL_LOW = "codelist"
CODE = "Code"
CODES = "Codes"
CODES_LOW = "codes"
CS = "ConceptScheme"
CS_LOW = "concept_scheme"
CON = "Concept"
CON_LOW = "concept"
DSD = "DataStructure"

# DSD components
DSD_COMPS = "DataStructureComponents"
CON_ID = "ConceptIdentity"
CON_ID_LOW = "concept_identity"
CON_ROLE = "ConceptRole"
CON_ROLE_LOW = "concept_role"
# Dimension
DIM_LIST = "DimensionList"
DIM_LIST_LOW = "dimension_list"
DIM = "Dimension"
TIME_DIM = "TimeDimension"
# Attribute
ATT_LIST = "AttributeList"
ATT_LIST_LOW = "attribute_list"
ATT = "Attribute"
ATT_LVL = "attachment_level"
ATT_REL_LOW = "attribute_relationship"
ATT_REL = "AttributeRelationship"
AS_STATUS = "assignmentStatus"
MANDATORY = "Mandatory"
CONDITIONAL = "Conditional"
# Measure
ME_LIST = "MeasureList"
ME_LIST_LOW = "measure_list"
MEASURE = "Measure"
PRIM_MEASURE = "PrimaryMeasure"
# Group Dimension
GROUP = "Group"
GROUP_DIM_LOW = "group_dimension_descriptor"
GROUP_DIM = "GroupDimension"
DIM_REF = "DimensionReference"

# Constraints
CON_CONS = "ContentConstraint"
CONS_ATT = "ConstraintAttachment"
CUBE_REGION = "CubeRegion"
CONTENT_REGION = "dataContentRegion"
KEY = "Key"
KEY_VALUE = "KeyValue"
DATA_KEY_SET = "DataKeySet"
DATA_KEY_SET_LOW = "dataKeySet"
INCLUDED = "isIncluded"
INCLUDE = "include"

# Annotation
ANNOTATION = "Annotation"
ANNOTATION_TITLE = "AnnotationTitle"
ANNOTATION_TYPE = "AnnotationType"
ANNOTATION_TEXT = "AnnotationText"
ANNOTATION_URL = "AnnotationURL"

TITLE = "title"
TYPE_ = "type_"
TYPE = "type"
TEXT = "text"
URL = "url"
URN = "URN"
URN_LOW = "urn"

# Representation
CORE_REP = "CoreRepresentation"
CORE_REP_LOW = "core_representation"
DTYPE = "dtype"
LOCAL_DTYPE = "local_dtype"
LOCAL_REP = "LocalRepresentation"
LOCAL_REP_LOW = "local_representation"
LOCAL_CODES = "LocalCodes"
LOCAL_CODES_LOW = "local_codes"
LOCAL_FACETS = "LocalFacets"
LOCAL_FACETS_LOW = "local_facets"
ENUM = "Enumeration"
ENUM_FORMAT = "EnumerationFormat"
TEXT_FORMAT = "TextFormat"

# Facets
FACETS = "facets"
TEXT_TYPE = "textType"
TEXT_TYPE_LOW = "text_type"

# Contact
CONTACT = "Contact"
CONTACTS_LOW = "contacts"
DEPARTMENT = "Department"
ROLE = "Role"
URI = "URI"
URI_LOW = "uri"
EMAIL = "Email"
X400 = "X400"
TELEPHONE = "Telephone"
FAX = "Fax"

# Extras
MAINTAINER = "maintainer"
XMLNS = "xmlns"
COMPS = "components"
PARENT = "Parent"
PAR_ID = "maintainableParentID"
PAR_VER = "maintainableParentVersion"
REL_TO = "relatedTo"
NO_REL = "NoSpecifiedRelationship"

# To exclude from attached_attributes
EXCLUDED_ATTRIBUTES = [STR_REF, "action", "dataScope", "xsi:type", SERIES, OBS]

# Content types
DATASTRUCTURES_CM = "DataStructures"
DATAFLOWS_CM = "Dataflows"
CODELISTS_CM = "Codelists"
CONCEPTS_CM = "Concepts"
ORGANISATIONSCHEMES_CM = "OrganisationSchemes"

VALID_FROM = "validFrom"
VALID_FROM_LOW = "valid_from"
VALID_TO = "validTo"
VALID_TO_LOW = "valid_to"

# Structures
IS_EXTERNAL_REF = "isExternalReference"
IS_EXTERNAL_REF_LOW = "is_external_reference"
IS_FINAL = "isFinal"
IS_FINAL_LOW = "is_final"
IS_PARTIAL = "isPartial"
IS_PARTIAL_LOW = "is_partial"

# Contact
URIS = "uris"
EMAILS = "emails"
TELEPHONES = "telephones"
FAXES = "faxes"

# Structure types
CLS = "Codelists"
DFW = "Dataflow"
DFWS = "Dataflows"
DFWS_LOW = "dataflows"
RULESETS = "Rulesets"
RULE_SCHEME = "RulesetScheme"
RULE = "Ruleset"
UDOS = "UserDefinedOperators"
UDO_SCHEME = "UserDefinedOperatorScheme"
UDO = "UserDefinedOperator"
TRANSFORMATIONS = "Transformations"
TRANS_SCHEME = "TransformationScheme"
TRANSFORMATION = "Transformation"



"""Model for VTL artefacts."""

from enum import Enum
from typing import Literal, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.errors import Invalid
from pysdmx.model.__base import Item, ItemScheme, Reference


class Transformation(Item, frozen=True, omit_defaults=True):
    """A statement which assigns the outcome of an expression to a result.

    Attributes:
        expression: str.
          The expression bound to the Transformation (no semicolon).
        result: str.
          The Dataset or Scalar name where we store the result.
        is_persistent: bool = False.
          If the result is persistent.
    """

    expression: str = ""
    is_persistent: bool = False
    result: str = ""

    @property
    def full_expression(self) -> str:
        """Return the full expression with the semicolon."""
        assign_operand = "<-" if self.is_persistent else ":="
        full_expression = f"{self.result} {assign_operand} {self.expression}"
        return (
            full_expression
            if full_expression.strip().endswith(";")
            else f"{full_expression};"
        )


class Ruleset(Item, frozen=True, omit_defaults=True):
    """A persistent set of rules."""

    ruleset_definition: str = ""
    ruleset_scope: Optional[Literal["variable", "valuedomain"]] = None
    ruleset_type: Optional[Literal["datapoint", "hierarchical"]] = None


class UserDefinedOperator(Item, frozen=True, omit_defaults=True):
    """Custom VTL operator that extends the VTL standard library."""

    operator_definition: str = ""


class NamePersonalisation(Item, frozen=True, omit_defaults=True):
    """Definition of personalised name."""

    personalised_name: str = ""
    vtl_artefact: str = ""
    vtl_default_name: str = ""


class VtlMapping(Item, frozen=True, omit_defaults=True):
    """Single VTL mapping."""


class ToVtlMapping(Struct, frozen=True, omit_defaults=True):
    """The mapping method and filter used when mapping from SDMX to VTL."""

    to_vtl_sub_space: Sequence[str] = ()
    method: Optional[str] = None


class FromVtlMapping(Struct, frozen=True, omit_defaults=True):
    """The mapping method and filter used when mapping from VTL to SDMX."""

    from_vtl_sub_space: Sequence[str] = ()
    method: Optional[str] = None


class VtlDataflowMapping(VtlMapping, frozen=True, omit_defaults=True):
    """Single mapping with a dataflow."""

    dataflow: str = ""
    dataflow_alias: str = ""
    to_vtl_mapping_method: Optional[ToVtlMapping] = None
    from_vtl_mapping_method: Optional[FromVtlMapping] = None


class VtlCodelistMapping(VtlMapping, frozen=True, omit_defaults=True):
    """Single mapping with a codelist."""

    codelist: str = ""
    codelist_alias: str = ""


class VtlConceptMapping(VtlMapping, frozen=True, omit_defaults=True):
    """Single mapping with a concept."""

    concept: str = ""
    concept_alias: str = ""


class CustomType(Item, frozen=True, omit_defaults=True):
    """Custom specification for a VTL basic scalar type."""

    data_type: str = ""
    null_value: Optional[str] = None
    output_format: Optional[str] = None
    vtl_literal_format: Optional[str] = None
    vtl_scalar_type: str = ""


class _VTLVersionEnum(Enum):
    """Enumeration of VTL versions."""

    VTL_1_0 = "1.0"
    VTL_2_0 = "2.0"
    VTL_2_1 = "2.1"


class VtlScheme(ItemScheme, frozen=True, omit_defaults=True):
    """A VTL item scheme with the additional propery 'vtl_version'."""

    vtl_version: Optional[str] = None

    def __post_init__(self) -> None:
        """Additional validation checks for VTL item schemes.

        Version must be a valid VTL version.

        Raises:
            Invalid: If the version is not a valid VTL version
        """
        if self.vtl_version not in _VTLVersionEnum._value2member_map_:
            raise Invalid(f"Invalid VTL version: {self.vtl_version}")


class CustomTypeScheme(VtlScheme, frozen=True, omit_defaults=True):
    """A collection of custom specifications for VTL basic scalar types."""


class NamePersonalisationScheme(VtlScheme, frozen=True, omit_defaults=True):
    """A collection of name personalisations."""


class RulesetScheme(VtlScheme, frozen=True, omit_defaults=True):
    """A collection of rulesets."""

    vtl_mapping_scheme: Optional[str] = None
    items: Sequence[Ruleset] = ()


class UserDefinedOperatorScheme(VtlScheme, frozen=True, omit_defaults=True):
    """A collection of user-defined operators."""

    vtl_mapping_scheme: Optional[str] = None
    ruleset_schemes: Sequence[Union[str, Reference]] = ()
    items: Sequence[UserDefinedOperator] = ()


class VtlMappingScheme(ItemScheme, frozen=True, omit_defaults=True):
    """A collection of VTL mappings."""


class TransformationScheme(VtlScheme, frozen=True, omit_defaults=True):
    """VTL Transformation Scheme class.

    TransformationScheme is a set of Transformations aimed at getting some
    meaningful results for the user
    (e.g. the validation of one or more Data Sets).

    This set of Transformations is meant to be executed together
    (in the same run) and may contain any number of Transformations to
    produce any number of results.
    Therefore, a TransformationScheme can be considered as a VTL program.

    The TransformationScheme must include the attribute vtlVersion expressed as
    a string (e.g. “2.0”), as the version of the VTL determines which syntax
    is used in defining the transformations of the scheme.
    This attribute is inherited from the VTLItemScheme class.

    Attributes:
        vtl_mapping_scheme: Optional[VtlMappingScheme].
          The VTL mapping scheme.
        name_personalisation_scheme: Optional[NamePersonalisationScheme].
          The name personalisation scheme.
        custom_type_scheme: Optional[CustomTypeScheme].
          The custom type scheme.
        ruleset_schemes: Sequence[RulesetScheme].
          The ruleset schemes.
        user_defined_operator_schemes: Sequence[UserDefinedOperatorScheme].
          The user-defined operator schemes.
    """

    vtl_mapping_scheme: Optional[VtlMappingScheme] = None
    name_personalisation_scheme: Optional[NamePersonalisationScheme] = None
    custom_type_scheme: Optional[CustomTypeScheme] = None
    ruleset_schemes: Sequence[Union[RulesetScheme, Reference]] = ()
    user_defined_operator_schemes: Sequence[
        Union[UserDefinedOperatorScheme, Reference]
    ] = ()
    items: Sequence[Transformation] = ()



"""SDMX All formats reader module."""

from io import BytesIO
from pathlib import Path
from typing import Optional, Sequence, Union

from pysdmx.errors import Invalid, NotFound
from pysdmx.io.format import Format
from pysdmx.io.input_processor import process_string_to_read
from pysdmx.model import Schema
from pysdmx.model.__base import ItemScheme
from pysdmx.model.dataflow import Dataflow, DataStructureDefinition
from pysdmx.model.dataset import Dataset
from pysdmx.model.message import Message
from pysdmx.model.submission import SubmissionResult
from pysdmx.util import parse_short_urn


def read_sdmx(
    sdmx_document: Union[str, Path, BytesIO],
    validate: bool = True,
) -> Message:
    """Reads any SDMX message and returns a dictionary.

    Supported structures formats are:
    - SDMX-ML 2.1 Structures

    Supported webservices submissions are:
    - SDMX-ML 2.1 RegistryInterface (Submission)
    - SDMX-ML 2.1 Error (raises an exception with the error content)

    Supported data formats are:
    - SDMX-ML 2.1
    - SDMX-CSV 1.0
    - SDMX-CSV 2.0

    Args:
        sdmx_document: Path to file (pathlib.Path), URL, or string.
        validate: Validate the input file (only for SDMX-ML).

    Returns:
        A dictionary containing the parsed SDMX data or metadata.

    Raises:
        Invalid: If the file is empty or the format is not supported.
    """
    input_str, read_format = process_string_to_read(sdmx_document)

    header = None
    result_data: Sequence[Dataset] = []
    result_structures: Sequence[
        Union[ItemScheme, Dataflow, DataStructureDefinition]
    ] = []
    result_submission: Sequence[SubmissionResult] = []
    if read_format == Format.STRUCTURE_SDMX_ML_2_1:
        from pysdmx.io.xml.sdmx21.reader.header import read as read_header
        from pysdmx.io.xml.sdmx21.reader.structure import (
            read as read_structure,
        )

        header = read_header(input_str, validate=validate)
        # SDMX-ML 2.1 Structure
        result_structures = read_structure(input_str, validate=validate)
    elif read_format == Format.DATA_SDMX_ML_2_1_GEN:
        from pysdmx.io.xml.sdmx21.reader.generic import read as read_generic
        from pysdmx.io.xml.sdmx21.reader.header import read as read_header

        header = read_header(input_str, validate=validate)
        # SDMX-ML 2.1 Generic Data
        result_data = read_generic(input_str, validate=validate)
    elif read_format == Format.DATA_SDMX_ML_2_1_STR:
        from pysdmx.io.xml.sdmx21.reader.header import read as read_header
        from pysdmx.io.xml.sdmx21.reader.structure_specific import (
            read as read_str_spe,
        )

        header = read_header(input_str, validate=validate)

        # SDMX-ML 2.1 Structure Specific Data
        result_data = read_str_spe(input_str, validate=validate)
    elif read_format == Format.REGISTRY_SDMX_ML_2_1:
        from pysdmx.io.xml.sdmx21.reader.submission import read as read_sub

        # SDMX-ML 2.1 Submission
        result_submission = read_sub(input_str, validate=validate)
    elif read_format == Format.ERROR_SDMX_ML_2_1:
        from pysdmx.io.xml.sdmx21.reader.error import read as read_error

        # SDMX-ML 2.1 Error
        read_error(input_str, validate=validate)
    elif read_format == Format.DATA_SDMX_CSV_1_0_0:
        from pysdmx.io.csv.sdmx10.reader import read as read_csv_v1

        # SDMX-CSV 1.0
        result_data = read_csv_v1(input_str)
    else:
        # SDMX-CSV 2.0
        from pysdmx.io.csv.sdmx20.reader import read as read_csv_v2

        result_data = read_csv_v2(input_str)

    if not (result_data or result_structures or result_submission):
        raise Invalid("Empty SDMX Message")

    # Returning a Message class
    if read_format in (
        Format.DATA_SDMX_CSV_1_0_0,
        Format.DATA_SDMX_CSV_2_0_0,
        Format.DATA_SDMX_ML_2_1_GEN,
        Format.DATA_SDMX_ML_2_1_STR,
    ):
        # TODO: Add here the Schema download for Datasets, based on structure
        # TODO: Ensure we have changed the signature of the data readers
        return Message(header=header, data=result_data)
    elif read_format == Format.REGISTRY_SDMX_ML_2_1:
        return Message(header=header, submission=result_submission)

    # TODO: Ensure we have changed the signature of the structure readers
    return Message(header=header, structures=result_structures)


def __assign_structure_to_dataset(
    datasets: Sequence[Dataset], structure_msg: Message
) -> None:
    for dataset in datasets:
        short_urn: str = (
            dataset.structure.short_urn
            if isinstance(dataset.structure, Schema)
            else dataset.structure
        )
        sdmx_type = parse_short_urn(short_urn).sdmx_type
        if sdmx_type == "DataStructure":
            try:
                dsd = structure_msg.get_data_structure_definition(short_urn)
                dataset.structure = dsd.to_schema()
            except NotFound:
                continue
        else:
            try:
                dataflow = structure_msg.get_dataflow(short_urn)
                dsd = structure_msg.get_data_structure_definition(
                    dataflow.structure if dataflow.structure else ""
                )
                dataset.structure = dsd.to_schema()
            except NotFound:
                continue


def get_datasets(
    data: Union[str, Path, BytesIO],
    structure: Optional[Union[str, Path, BytesIO]] = None,
    validate: bool = True,
) -> Sequence[Dataset]:
    """Reads a data message and a structure message and returns a dataset.

    Args:
        data: Path to file (pathlib.Path), URL, or string for the data message.
        structure:
          Path to file (pathlib.Path), URL, or string
          for the structure message, if needed.
        validate: Validate the input file (only for SDMX-ML).

    Returns:
        A sequence of Datasets

    Raises:
        Invalid:
            If the data message is empty or the related data structure
            (or dataflow with its children) is not found.
        NotFound:
            If the related data structure (or dataflow with its children)
            is not found.
    """
    data_msg = read_sdmx(data, validate=validate)
    if not data_msg.data:
        raise Invalid("No data found in the data message")

    if structure is None:
        return data_msg.data
    structure_msg = read_sdmx(structure, validate=validate)
    if structure_msg.structures is None:
        raise Invalid("No structure found in the structure message")

    __assign_structure_to_dataset(data_msg.data, structure_msg)

    # Check if any dataset does not have a structure
    for dataset in data_msg.data:
        if not isinstance(dataset.structure, Schema):
            raise Invalid(
                f"Missing DataStructure for dataset {dataset.short_urn}"
            )

    return data_msg.data



"""Model for SDMX Categories.

An example of a category scheme is one which categorises data, sometimes
known as a subject matter domain scheme or a data category scheme.
"""

from typing import Iterator, Optional, Sequence, Union

from pysdmx.model.__base import (
    DataflowRef,
    Item,
    ItemScheme,
    MaintainableArtefact,
)
from pysdmx.model.dataflow import Dataflow


class Category(Item, frozen=False, omit_defaults=True):  # type: ignore[misc]
    """A category, ie a way to **organize and group** things.

    Categories are used to organize and group other artefacts in SDMX.
    A typical example would be to organize dataflows by topics
    (e.g. derivatives, exchange rates, etc.)

    Categories can be **nested**, up to any level of depth and,
    therefore, a category is **iterable**.

    Attributes:
        id: The category ID (e.g. TOPIC_001)
        name: The category name (e.g. "Locational banking statistics")
        description: Additionale descriptive information about the
            category (e.g. "These statistics cover the balance sheets
            of internationally active banks.")
        categories: The sub-categories, i.e. the categories belonging
            to the category.
        dataflows: The list of dataflows attached to the category.
    """

    categories: Sequence["Category"] = ()
    dataflows: Sequence[Union[Dataflow, DataflowRef]] = ()

    def __iter__(self) -> Iterator["Category"]:
        """Return an iterator over the list of categories."""
        yield from self.categories


class CategoryScheme(ItemScheme, frozen=True, omit_defaults=True):
    """An immutable collection of categories, likes a list of topics.

    A category scheme is **maintained by its agency**, typically, an
    organisation like the BIS, the ECB, the IMF, SDMX, etc.

    A category scheme has an **identifier** and a **name**. It may also
    have a **description** and **version**.

    A category scheme is **iterable**.

    While every category contains the list of dataflows attached to it,
    the ``dataflows`` property of the ``CategoryScheme`` can be used
    to retrieve (as a set) the list of dataflows attached to any of the
    categories in the scheme.

    Attributes:
        id: The identifier for the scheme (e.g. STAT_SUBJECT_MATTER).
        name: The scheme name (e.g. "SDMX Statistical Subject-Matter
            Domains").
        agency: The maintainer of the scheme (e.g. SDMX).
        description: Additional descriptive information about the scheme
            (e.g. "The SDMX Content Guidelines for Statistical
            Subject-Matter Domains").
        version: The scheme version (e.g. 1.0)
    """

    items: Sequence[Category] = ()

    @property
    def categories(self) -> Sequence[Category]:
        """The list of top level categories in the scheme."""
        return self.items

    @property
    def dataflows(self) -> Union[Sequence[DataflowRef], Sequence[Dataflow]]:
        """Return the dataflows attached to any category in the scheme."""
        flows = set()  # type: ignore[var-annotated]
        for cat in self.categories:
            flows.update(self.__extract_flows(cat))
        return list(flows)

    def __iter__(self) -> Iterator[Category]:
        """Return an iterator over the list of categories."""
        yield from self.categories

    def __len__(self) -> int:
        """Return the number of categories in the category scheme."""
        return self.__get_count(self.categories)

    def __getitem__(self, id_: str) -> Optional[Category]:
        """Return the category identified by the given ID."""
        return self.__extract_cat(self.categories, id_)

    def __contains__(self, id_: str) -> bool:
        """Whether there is a category with the supplied ID in the scheme."""
        return bool(self.__getitem__(id_))

    def __get_count(self, categories: Sequence[Category]) -> int:
        """Return the number of categories at all levels."""
        count = len(categories)
        for cat in categories:
            if cat.categories:
                count += self.__get_count(cat.categories)
        return count

    def __extract_cat(
        self, categories: Sequence[Category], id_: str
    ) -> Optional[Category]:
        if "." in id_:
            ids = id_.split(".")
            out = list(filter(lambda cat: cat.id == ids[0], categories))
            if out:
                pkey = ".".join(ids[1:])
                return self.__extract_cat(out[0].categories, pkey)
        else:
            out = list(filter(lambda cat: cat.id == id_, categories))
            if out:
                return out[0]
        return None

    def __extract_flows(
        self, c: Category
    ) -> Union[Sequence[DataflowRef], Sequence[Dataflow]]:
        flows = []  # type: ignore[var-annotated]
        if c.dataflows:
            flows.extend(c.dataflows)
        for sub in c.categories:
            flows.extend(self.__extract_flows(sub))
        return flows


class Categorisation(
    MaintainableArtefact, frozen=True, omit_defaults=True, kw_only=True
):
    """Link between a category and an artefact attached to it."""

    source: str
    target: str



"""Message module.

This module contains the enumeration for the different types of messages that
can be written. It also contains the Header and Message classes that are used
to create the SDMX messages.

Classes:
    Header: Header for the SDMX messages.
    ActionType: Enumeration for the different types of actions that can be
        written.
    Message: Class that holds the content of the SDMX message.
    SubmissionResult: Class that represents the result of a submission.
"""

import uuid
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional, Sequence, Type, Union

from msgspec import Struct

from pysdmx.errors import Invalid, NotFound
from pysdmx.model import (
    AgencyScheme,
    RulesetScheme,
    TransformationScheme,
    UserDefinedOperatorScheme,
)
from pysdmx.model.__base import ItemScheme, Organisation
from pysdmx.model.code import Codelist
from pysdmx.model.concept import ConceptScheme
from pysdmx.model.dataflow import Dataflow, DataStructureDefinition
from pysdmx.model.dataset import ActionType, Dataset
from pysdmx.model.submission import SubmissionResult


class Header(Struct, kw_only=True):
    """Header for the SDMX messages."""

    id: str = str(uuid.uuid4())
    test: bool = False
    prepared: datetime = datetime.now(timezone.utc)
    sender: Organisation = Organisation(id="ZZZ")
    receiver: Optional[Organisation] = None
    source: Optional[str] = None
    dataset_action: Optional[ActionType] = None
    structure: Optional[Dict[str, str]] = None
    dataset_id: Optional[str] = None


class Message(Struct, frozen=True):
    """Message class holds the content of SDMX Message.

    Attributes:
        structures: Sequence of structure objects (ItemScheme, Schema).
           They represent the contents of a Structure Message.
        data: Sequence of Dataset objects. They represent the contents of a
           SDMX Data Message in any format.
        submission: Sequence of SubmissionResult objects. They represent the
              contents of a SDMX Submission Message.
    """

    header: Optional[Header] = None
    structures: Optional[
        Sequence[
            Union[
                ItemScheme,
                DataStructureDefinition,
                Dataflow,
            ]
        ]
    ] = None
    data: Optional[Sequence[Dataset]] = None
    submission: Optional[Sequence[SubmissionResult]] = None

    def __post_init__(self) -> None:
        """Checks if the content is valid."""
        if self.structures is not None:
            for obj_ in self.structures:
                if not isinstance(
                    obj_, (ItemScheme, DataStructureDefinition, Dataflow)
                ):
                    raise Invalid(
                        f"Invalid structure: " f"{type(obj_).__name__} ",
                        "Check the docs on structures.",
                    )
        if self.data is not None:
            for data_value in self.data:
                if not isinstance(data_value, Dataset):
                    raise Invalid(
                        f"Invalid data type: "
                        f"{type(data_value).__name__} "
                        f"for Data Message, requires a Dataset object.",
                        "Check the docs for the proper structure on data.",
                    )

    def __get_elements(self, type_: Type[Any]) -> List[Any]:
        """Returns a list of elements of a specific type."""
        if self.structures is None:
            raise NotFound(
                f"No {type_.__name__} found in message.",
            )
        structures = []
        for element in self.structures:
            if isinstance(element, type_):
                structures.append(element)
        return structures

    def __get_single_structure(
        self,
        type_: Type[Union[ItemScheme, DataStructureDefinition, Dataflow]],
        short_urn: str,
    ) -> Any:
        """Returns a specific element from content."""
        if self.structures is None:
            raise NotFound(
                f"No {type_.__name__} found in message.",
                "Could not find any Structures in this message.",
            )
        for structure in self.structures:
            if structure.short_urn == short_urn:
                return structure

        raise NotFound(
            f"No {type_.__name__} with Short URN {short_urn} found in message",
            "Could not find the requested element.",
        )

    def get_agency_schemes(self) -> List[AgencyScheme]:
        """Returns the AgencySchemes."""
        return self.__get_elements(AgencyScheme)

    def get_codelists(self) -> List[Codelist]:
        """Returns the Codelists."""
        return self.__get_elements(Codelist)

    def get_concept_schemes(self) -> List[ConceptScheme]:
        """Returns the Concept Schemes."""
        return self.__get_elements(ConceptScheme)

    def get_data_structure_definitions(
        self,
    ) -> List[DataStructureDefinition]:
        """Returns the DataStructureDefinitions."""
        return self.__get_elements(DataStructureDefinition)

    def get_dataflows(self) -> List[Dataflow]:
        """Returns the Dataflows."""
        return self.__get_elements(Dataflow)

    def get_organisation_scheme(self, short_urn: str) -> AgencyScheme:
        """Returns a specific OrganisationScheme."""
        return self.__get_single_structure(AgencyScheme, short_urn)

    def get_codelist(self, short_urn: str) -> Codelist:
        """Returns a specific Codelist."""
        return self.__get_single_structure(Codelist, short_urn)

    def get_concept_scheme(self, short_urn: str) -> ConceptScheme:
        """Returns a specific Concept Scheme."""
        return self.__get_single_structure(ConceptScheme, short_urn)

    def get_data_structure_definition(
        self, short_urn: str
    ) -> DataStructureDefinition:
        """Returns a specific DataStructureDefinition."""
        return self.__get_single_structure(DataStructureDefinition, short_urn)

    def get_dataflow(self, short_urn: str) -> Dataflow:
        """Returns a specific Dataflow."""
        return self.__get_single_structure(Dataflow, short_urn)

    def get_datasets(self) -> Sequence[Dataset]:
        """Returns the Datasets."""
        if self.data is not None:
            return self.data
        raise NotFound(
            "No Datasets found in data.",
            "Could not find any Datasets in content.",
        )

    def get_dataset(self, short_urn: str) -> Dataset:
        """Returns a specific Dataset."""
        if self.data is not None:
            for dataset in self.data:
                if dataset.short_urn == short_urn:
                    return dataset
        raise NotFound(
            f"No Dataset with Short URN {short_urn} found in data.",
            "Could not find the requested Dataset.",
        )

    def get_transformation_schemes(self) -> List[TransformationScheme]:
        """Returns the TransformationSchemes."""
        return self.__get_elements(TransformationScheme)

    def get_user_defined_operator_schemes(
        self,
    ) -> List[UserDefinedOperatorScheme]:
        """Returns the UserDefinedOperatorSchemes."""
        return self.__get_elements(UserDefinedOperatorScheme)

    def get_ruleset_schemes(self) -> List[RulesetScheme]:
        """Returns the RulesetSchemes."""
        return self.__get_elements(RulesetScheme)



"""Model validations for VTL models."""

from typing import Union

from pysdmx.__extras_check import __check_vtl_extra
from pysdmx.errors import Invalid
from pysdmx.model import (
    Ruleset,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlScheme,
)
from pysdmx.model.__base import Item
from pysdmx.toolkit.vtl._validations import (
    _ruleset_scheme_validations,
    _ruleset_validation,
    _transformation_scheme_validations,
    _transformation_validations,
    _user_defined_operator_scheme_validations,
    _user_defined_operator_validation,
)


def model_validations(model_obj: Union[VtlScheme, Item]) -> None:
    """Validation checks for VTL models.

    Args:
        model_obj: A vtlscheme or Item object.
    raises:
        Invalid: Invalid model object if the model object is not valid.
    """
    __check_vtl_extra()

    if isinstance(model_obj, Ruleset):
        _ruleset_validation(model_obj)
    elif isinstance(model_obj, RulesetScheme):
        _ruleset_scheme_validations(model_obj)
    elif isinstance(model_obj, UserDefinedOperator):
        _user_defined_operator_validation(model_obj)
    elif isinstance(model_obj, UserDefinedOperatorScheme):
        _user_defined_operator_scheme_validations(model_obj)
    elif isinstance(model_obj, Transformation):
        _transformation_validations(model_obj)
    elif isinstance(model_obj, TransformationScheme):
        _transformation_scheme_validations(model_obj)
    else:
        raise Invalid("Invalid model object")



from typing import Dict, List, Optional, Sequence, Tuple

import pandas as pd

from pysdmx.errors import Invalid
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.sdmx21.writer.__write_aux import ALL_DIM
from pysdmx.model import Role, Schema


def check_content_dataset(content: Sequence[PandasDataset]) -> None:
    """Checks if the Message content is a dataset."""
    if not all(isinstance(dataset, PandasDataset) for dataset in content):
        raise Invalid("Message Content must only contain a Dataset sequence.")


def check_dimension_at_observation(
    datasets: Dict[str, PandasDataset],
    dimension_at_observation: Optional[Dict[str, str]],
) -> Dict[str, str]:
    """This function checks if the dimension at observation is valid."""
    # If dimension_at_observation is None, set it to ALL_DIM
    if dimension_at_observation is None:
        dimension_at_observation = dict.fromkeys(datasets, ALL_DIM)
        return dimension_at_observation

    # Check the datasets and their dimensions are present
    for key, value in dimension_at_observation.items():
        if key not in datasets:
            raise Invalid(f"Dataset {key} not found in Message content.")
        writing_validation(datasets[key])
        dataset = datasets[key]
        components = dataset.structure.components  # type: ignore[union-attr]

        dimension_codes = [dim.id for dim in components.dimensions]
        if value not in dimension_codes:
            raise Invalid(
                f"Dimension at observation {value} "
                f"not found in dataset {key}."
            )
    # Add the missing datasets on mapping with ALL_DIM
    for key in datasets:
        if key not in dimension_at_observation:
            dimension_at_observation[key] = ALL_DIM
    return dimension_at_observation


def writing_validation(dataset: PandasDataset) -> None:
    """Structural validation of the dataset."""
    if not isinstance(dataset.structure, Schema):
        raise Invalid(
            "Dataset Structure is not a Schema. Cannot perform operation."
        )
    required_components = [
        comp.id
        for comp in dataset.structure.components
        if comp.role in (Role.DIMENSION, Role.MEASURE)
    ]
    for att in dataset.structure.components.attributes:
        if (
            att.required
            and att.attachment_level is not None
            and att.attachment_level != "D"
        ):
            required_components.append(att.id)
    non_required = [
        comp.id
        for comp in dataset.structure.components
        if comp.id not in required_components
    ]
    # Columns match components
    columns = dataset.data.columns
    all_components = required_components + non_required
    difference = [col for col in columns if col not in all_components]

    for comp in required_components:
        difference.append(comp) if comp not in columns else None
    if difference:
        raise Invalid(
            f"Data columns must match components. "
            f"Difference: {', '.join(difference)}"
        )
    # Check if the dataset has at least one dimension and one measure
    if not dataset.structure.components.dimensions:
        raise Invalid(
            "The dataset structure must have at least one dimension."
        )
    if not dataset.structure.components.measures:
        raise Invalid("The dataset structure must have at least one measure.")


def get_codes(
    dimension_code: str, structure: Schema, data: pd.DataFrame
) -> Tuple[List[str], List[str]]:
    """This function divides the components in Series and Obs."""
    series_codes = []
    obs_codes = [dimension_code, structure.components.measures[0].id]

    # Getting the series and obs codes
    for dim in structure.components.dimensions:
        if dim.id != dimension_code:
            series_codes.append(dim.id)

    # Adding the attributes based on the attachment level
    for att in structure.components.attributes:
        if att.attachment_level == "O" and att.id in data.columns:
            obs_codes.append(att.id)
        elif (
            att.attachment_level is not None
            and att.attachment_level != "D"
            and att.id in data.columns
        ):
            series_codes.append(att.id)

    return series_codes, obs_codes



"""Read SDMX-ML submission messages."""

from typing import Any, Dict, Sequence

from pysdmx.errors import Invalid
from pysdmx.io.xml.sdmx21.__tokens import (
    ACTION,
    MAINTAINABLE_OBJECT,
    REG_INTERFACE,
    STATUS,
    STATUS_MSG,
    SUBMISSION_RESULT,
    SUBMIT_STRUCTURE_RESPONSE,
    SUBMITTED_STRUCTURE,
    URN,
)
from pysdmx.io.xml.sdmx21.reader.__parse_xml import parse_xml
from pysdmx.model.submission import SubmissionResult
from pysdmx.util import parse_urn


def __handle_registry_interface(
    dict_info: Dict[str, Any],
) -> Sequence[SubmissionResult]:
    """Handle the Registry Interface message.

    Args:
        dict_info: Dictionary with the parsed data.

    Returns:
        dict: Dictionary with the parsed data.
    """
    response = dict_info[REG_INTERFACE][SUBMIT_STRUCTURE_RESPONSE]

    result = []
    for submission_result in response[SUBMISSION_RESULT]:
        structure = submission_result[SUBMITTED_STRUCTURE]
        action = structure[ACTION]
        urn = structure[MAINTAINABLE_OBJECT][URN]
        short_urn = str(parse_urn(urn))
        status = submission_result[STATUS_MSG][STATUS]
        sr = SubmissionResult(action, short_urn, status)
        result.append(sr)
    return result


def read(input_str: str, validate: bool = True) -> Sequence[SubmissionResult]:
    """Reads an SDMX-ML 2.1 Submission Result file.

    Args:
        input_str: SDMX-ML data to read.
        validate: If True, the XML data will be validated against the XSD.
    """
    dict_info = parse_xml(input_str, validate=validate)
    if REG_INTERFACE not in dict_info:
        raise Invalid("This SDMX document is not an SDMX-ML 2.1 Submission.")
    return __handle_registry_interface(dict_info)



"""Collection of Fusion-JSON schemas for common artefacts."""

from datetime import datetime
from typing import Any, Optional, Sequence, Union

import msgspec

from pysdmx.errors import NotFound
from pysdmx.model import ArrayBoundaries, Codelist, Facets
from pysdmx.util import find_by_urn


class FusionString(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for an international string."""

    locale: str
    value: str


class FusionAnnotation(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for annotations."""

    id: Optional[str] = None
    title: Optional[str] = None
    type: Optional[str] = None
    value: Optional[str] = None
    text: Sequence[FusionString] = ()


class FusionLink(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for link objects."""

    urn: str


class FusionTextFormat(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for TextFormat."""

    textType: str
    minLength: Optional[int] = None
    maxLength: Optional[int] = None
    minValue: Optional[str] = None
    maxValue: Optional[str] = None
    startValue: Optional[str] = None
    endValue: Optional[str] = None
    decimals: Optional[int] = None
    pattern: Optional[str] = None
    startTime: Optional[datetime] = None
    endTime: Optional[datetime] = None
    isSequence: bool = False


class FusionRepresentation(msgspec.Struct, frozen=True):
    """Fusion-JSON payload for core representation."""

    textFormat: Optional[FusionTextFormat] = None
    representation: Optional[str] = None
    minOccurs: Optional[int] = None
    maxOccurs: Optional[int] = None

    def __to_number(self, input: str) -> Union[int, float]:
        try:
            return int(input)
        except ValueError:
            return float(input)

    def to_facets(self) -> Optional[Facets]:
        """Return a Facets domain object."""
        if self.textFormat and (
            self.textFormat.minLength
            or self.textFormat.maxLength
            or self.textFormat.isSequence
            or self.textFormat.minValue
            or self.textFormat.maxValue
            or self.textFormat.startValue
            or self.textFormat.endValue
            or self.textFormat.decimals
            or self.textFormat.pattern
            or self.textFormat.startTime
            or self.textFormat.endTime
        ):
            min = (
                self.__to_number(self.textFormat.minValue)
                if self.textFormat.minValue
                else None
            )
            max = (
                self.__to_number(self.textFormat.maxValue)
                if self.textFormat.maxValue
                else None
            )
            sta = (
                self.__to_number(self.textFormat.startValue)
                if self.textFormat.startValue
                else None
            )
            end = (
                self.__to_number(self.textFormat.endValue)
                if self.textFormat.endValue
                else None
            )

            return Facets(
                min_length=self.textFormat.minLength,
                max_length=self.textFormat.maxLength,
                is_sequence=self.textFormat.isSequence,
                min_value=min,
                max_value=max,
                start_value=sta,
                end_value=end,
                decimals=self.textFormat.decimals,
                pattern=self.textFormat.pattern,
                start_time=self.textFormat.startTime,
                end_time=self.textFormat.endTime,
            )
        else:
            return None

    def to_enumeration(
        self,
        codelists: Sequence[Any],
        valid: Sequence[str],
    ) -> Optional[Codelist]:
        """Returns the list of codes allowed for this component."""
        if self.representation:
            try:
                a = find_by_urn(codelists, self.representation)
                cl = a.to_model()
                codes = [
                    c.to_model() for c in a.items if not valid or c.id in valid
                ]
                return msgspec.structs.replace(cl, items=codes)
            except NotFound:
                # This is OK. In case of schema queries, if a component
                # has both a local and a core representations, only the
                # relevant one (i.e. the local one) will be included in
                # the schema.
                return None
        return None

    def to_array_def(self) -> Optional[ArrayBoundaries]:
        """Returns the array boundaries, if any."""
        if self.maxOccurs and self.maxOccurs > 1:
            m = self.minOccurs if self.minOccurs is not None else 0
            return ArrayBoundaries(m, self.maxOccurs)
        else:
            return None



"""A module to represent a Submission Result."""

from msgspec import Struct


class SubmissionResult(Struct, frozen=True):
    """A class to represent a Submission Result."""

    action: str
    short_urn: str
    status: str

    def __str__(self) -> str:
        """Return a string representation of the SubmissionResult."""
        return (
            f"<Submission Result - "
            f"Action: {self.action} - "
            f"Short URN: {self.short_urn} - "
            f"Status: {self.status}>"
        )



"""SDMX 1.0 CSV reader module."""

from io import StringIO
from typing import Sequence

import pandas as pd

from pysdmx.errors import Invalid
from pysdmx.io.pd import PandasDataset


def __generate_dataset_from_sdmx_csv(data: pd.DataFrame) -> PandasDataset:
    # For SDMX-CSV version 1, use 'DATAFLOW' column as the structure id
    structure_id = data["DATAFLOW"].iloc[0]
    # Drop 'DATAFLOW' column from DataFrame
    df_csv = data.drop(["DATAFLOW"], axis=1)
    urn = f"Dataflow={structure_id}"

    # Return a Dataset object with the extracted information
    return PandasDataset(
        structure=urn,
        data=df_csv,
    )


def read(input_str: str) -> Sequence[PandasDataset]:
    """Reads csv data and returns a sequence of Datasets.

    Args:
        input_str: str.

    Returns:
        A Sequence of Pandas Datasets.

    Raises:
        Invalid: If it is an invalid CSV file.
    """
    # Get Dataframe from CSV file
    df_csv = pd.read_csv(StringIO(input_str))
    # Drop empty columns
    df_csv = df_csv.dropna(axis=1, how="all")

    # Determine SDMX-CSV version based on column names
    if "DATAFLOW" not in df_csv.columns:
        # Raise an exception if the CSV file is not in SDMX-CSV format
        raise Invalid(
            "Only SDMX-CSV 1.0 is allowed",
            "Invalid SDMX-CSV 1.0 file. "
            "Check the docs for the proper structure on content.",
        )

    # Convert all columns to strings
    df_csv = df_csv.astype("str")
    # Check if any column headers contain ':', indicating mode, label or text
    mode_label_text = any(":" in x for x in df_csv.columns)

    # Determine the id column based on the SDMX-CSV version
    id_column = "DATAFLOW"

    # If mode, label or text is present, modify the DataFrame
    if mode_label_text:
        # Split the ID column to remove mode, label or text
        df_csv[id_column] = df_csv[id_column].map(lambda x: x.split(": ")[0])
        # Split the other columns to remove mode, label, or text
        sequence = 1
        for x in df_csv.columns[sequence:]:
            df_csv[x.split(":")[0]] = df_csv[x].map(
                lambda x: x.split(": ", 2)[0], na_action="ignore"
            )
            # Delete the original columns
            del df_csv[x]

    # Separate SDMX-CSV in different datasets per Structure ID
    list_df = [data for _, data in df_csv.groupby(id_column)]

    # Create a payload dictionary to store datasets with the
    # different unique_ids as keys
    payload = []
    for df in list_df:
        # Generate a dataset from each subset of the DataFrame
        dataset = __generate_dataset_from_sdmx_csv(data=df)

        # Add the dataset to the payload dictionary
        payload.append(dataset)

    # Return the payload generated
    return payload



"""Types of errors that can be returned by pysdmx."""

from typing import Any, Dict, Optional


class PysdmxError(Exception):
    """Base ``pysdmx`` error class.

    This class is inherited by more specific error classes and so, ideally,
    you should not encounter it directly when calling library functions.

    Attributes:
        title: A short description of the problem.
        description: A longer, human-friendly description of the problem,
            ideally with tips on how the issue can be resolved.
        csi: A dictionary containing additional details about the issue,
            that can contribute to its understanding or resolution.
    """

    def __init__(
        self,
        title: str,
        description: Optional[str] = None,
        csi: Optional[Dict[str, Any]] = None,
    ):
        """Instantiates a new error, with the supplied paramaters."""
        self.title = title
        self.description = description
        self.csi = csi
        msg = f"{title}: {description}" if description else title
        super().__init__(msg)


class RetriableError(PysdmxError):
    """A type of errors that may be resolved after retrying.

    This class is inherited by more specific error classes and so, ideally,
    you should not encounter it directly when calling library functions.
    """


class Unavailable(RetriableError):
    """The targeted service or resource is not available.

    This can be for a variety of reasons such as:

    - A network issue between the client making the call and the targeted
      service
    - The targeted service is temporarily unavailable
    - The targeted service is overloaded and therefore unable to process
      the request

    This type of errors is considered as **retriable** and so clients can
    retry executing the query at a later stage.
    """


class Invalid(PysdmxError):
    """The request is invalid.

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating why the request
    is invalid.
    """


class InternalError(PysdmxError):
    """The request is valid but could not be fulfilled.

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating (and possibly
    reporting) the issue first.
    """


class NotFound(PysdmxError):
    """The requested resource is **not available**.

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating the issue first.
    """


class NotImplemented(PysdmxError):
    """The requested operation is not supported.

    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating the issue first.
    """


class Unauthorized(PysdmxError):
    """The request was **not authorized**.

    This can be for a variety of reasons such as:

    - Authentication is required by no credentials were supplied.
    - Credentials were supplied but were invalid.
    - Credentials were supplied and valid but access is denied.


    This type of errors is considered as **non-retriable** and so clients
    should **not** retry the query before investigating the issue first.
    """



"""Collection of Fusion-JSON schemas for structure map queries."""

from datetime import datetime as dt
from datetime import timezone as tz
from typing import Any, Dict, Optional, Sequence, Union

from msgspec import Struct

from pysdmx.io.json.fusion.messages.core import FusionString
from pysdmx.model import (
    ComponentMap,
    DataType,
    DatePatternMap,
    FixedValueMap,
    ImplicitComponentMap,
    MultiComponentMap,
    MultiRepresentationMap,
    MultiValueMap,
    RepresentationMap,
    ValueMap,
)
from pysdmx.model import (
    StructureMap as SM,
)
from pysdmx.util import find_by_urn


class FusionSourceValue(Struct, frozen=True):
    """Fusion-JSON payload for a source value."""

    value: str
    regEx: bool = False

    def to_model(self) -> str:
        """Returns the requested source value."""
        if self.regEx:
            return f"regex:{self.value}"
        else:
            return self.value


class FusionRepresentationMapping(Struct, frozen=True):
    """Fusion-JSON payload for a representation mapping."""

    source: Sequence[FusionSourceValue]
    target: Sequence[str]
    validFrom: Optional[str] = None
    validTo: Optional[str] = None

    def __get_dt(self, inp: str) -> dt:
        if inp.endswith("Z"):
            inp = inp[0:-1]
        return dt.fromisoformat(inp).replace(tzinfo=tz.utc)

    def to_model(self, is_multi: bool) -> Union[MultiValueMap, ValueMap]:
        """Returns the requested value maps."""
        if is_multi:
            return MultiValueMap(
                source=[src.to_model() for src in self.source],
                target=self.target,
                valid_from=(
                    self.__get_dt(self.validFrom) if self.validFrom else None
                ),
                valid_to=self.__get_dt(self.validTo) if self.validTo else None,
            )
        else:
            return ValueMap(
                source=self.source[0].to_model(),
                target=self.target[0],
                valid_from=(
                    self.__get_dt(self.validFrom) if self.validFrom else None
                ),
                valid_to=self.__get_dt(self.validTo) if self.validTo else None,
            )


class FusionRepresentationMap(
    Struct,
    frozen=True,
    rename={"agency": "agencyId"},
):
    """Fusion-JSON payload for a representation map."""

    id: str
    names: Sequence[FusionString]
    agency: str
    version: str
    sources: Sequence[str]
    targets: Sequence[str]
    mappedRelationships: Sequence[FusionRepresentationMapping]
    descriptions: Sequence[FusionString] = ()

    def to_model(
        self, is_multi: bool = False
    ) -> Union[MultiRepresentationMap, RepresentationMap]:
        """Returns the requested representation map."""
        mrs = [rm.to_model(is_multi) for rm in self.mappedRelationships]
        s = [i if i.startswith("urn:") else DataType(i) for i in self.sources]
        t = [j if j.startswith("urn:") else DataType(j) for j in self.targets]
        if is_multi:
            return MultiRepresentationMap(
                id=self.id,
                name=self.names[0].value,
                agency=self.agency,
                source=s,
                target=t,
                maps=mrs,  # type: ignore[arg-type]
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                version=self.version,
            )
        else:
            return RepresentationMap(
                id=self.id,
                name=self.names[0].value,
                agency=self.agency,
                source=s[0],
                target=t[0],
                maps=mrs,  # type: ignore[arg-type]
                description=(
                    self.descriptions[0].value if self.descriptions else None
                ),
                version=self.version,
            )


class FusionComponentMap(Struct, frozen=True):
    """Fusion-JSON payload for a component map."""

    sources: Sequence[str]
    targets: Sequence[str]
    representationMapRef: Optional[str] = None

    def to_model(
        self, rms: Sequence[FusionRepresentationMap]
    ) -> Union[ComponentMap, MultiComponentMap, ImplicitComponentMap]:
        """Returns the requested component map."""
        if self.representationMapRef:
            rm = find_by_urn(rms, self.representationMapRef)
            if len(self.sources) == 1 and len(self.targets) == 1:
                return ComponentMap(
                    self.sources[0],
                    self.targets[0],
                    rm.to_model(),
                )
            else:
                return MultiComponentMap(
                    self.sources, self.targets, rm.to_model(True)
                )
        else:
            return ImplicitComponentMap(self.sources[0], self.targets[0])


class FusionTimePatternMap(Struct, frozen=True):
    """Fusion-JSON payload for a date pattern map."""

    source: str
    target: str
    pattern: str
    locale: str
    freqId: Optional[str] = None
    freqDim: Optional[str] = None
    id: Optional[str] = None

    def to_model(self) -> DatePatternMap:
        """Returns the requested date mapper."""
        freq = self.freqId if self.freqId else self.freqDim
        typ = "fixed" if self.freqId else "variable"
        return DatePatternMap(
            self.source,
            self.target,
            self.pattern,
            freq,  # type: ignore[arg-type]
            self.id,
            self.locale,
            typ,  # type: ignore[arg-type]
        )


class FusionStructureMap(Struct, frozen=True):
    """Fusion-JSON payload for a structure map."""

    id: str
    agencyId: str
    version: str
    source: str
    target: str
    names: Sequence[FusionString] = ()
    descriptions: Sequence[FusionString] = ()
    fixedInput: Dict[str, Any] = {}
    fixedOutput: Dict[str, Any] = {}
    timePatternMaps: Sequence[FusionTimePatternMap] = ()
    componentMaps: Sequence[FusionComponentMap] = ()

    def to_model(
        self,
        rms: Sequence[FusionRepresentationMap],
    ) -> SM:
        """Returns the requested mapping definition."""
        m1 = tuple(tpm.to_model() for tpm in self.timePatternMaps)
        m2 = tuple(cm.to_model(rms) for cm in self.componentMaps)
        m3 = tuple(FixedValueMap(k, v) for k, v in self.fixedOutput.items())
        m4 = tuple(
            FixedValueMap(k, v, "source") for k, v in self.fixedInput.items()
        )

        return SM(
            id=self.id,
            name=self.names[0].value,
            agency=self.agencyId,
            source=self.source,
            target=self.target,
            maps=m1 + m2 + m3 + m4,
            description=(
                self.descriptions[0].value if self.descriptions else None
            ),
            version=self.version,
        )


class FusionMappingMessage(Struct, frozen=True):
    """Fusion-JSON payload for /structuremap queries."""

    StructureMap: Sequence[FusionStructureMap]
    RepresentationMap: Sequence[FusionRepresentationMap] = ()

    def to_model(self) -> SM:
        """Returns the requested mapping definition."""
        return self.StructureMap[0].to_model(self.RepresentationMap)


class FusionRepresentationMapMessage(Struct, frozen=True):
    """Fusion-JSON payload for /representationmap queries."""

    RepresentationMap: Sequence[FusionRepresentationMap]

    def to_model(self) -> SM:
        """Returns the requested mapping definition."""
        m = self.RepresentationMap[0]
        multi = bool(len(m.sources) > 1 or len(m.targets) > 1)
        out = m.to_model(multi)
        return out  # type: ignore[return-value]



"""Model for SDMX agency schemes and data provider schemes."""

from typing import Iterator, Optional, Sequence

from pysdmx.model.__base import (
    Agency,
    DataConsumer,
    DataProvider,
    ItemScheme,
    MetadataProvider,
)


class AgencyScheme(ItemScheme, frozen=True, omit_defaults=True):
    """An immutable collection of agencies."""

    id: str = "AGENCIES"
    name: str = "AGENCIES"
    version: str = "1.0"
    items: Sequence[Agency] = ()

    @property
    def agencies(self) -> Sequence[Agency]:
        """Extract the items in the scheme."""
        return self.items

    def __iter__(self) -> Iterator[Agency]:
        """Return an iterator over the list of agencies."""
        yield from self.items

    def __len__(self) -> int:
        """Return the number of agencies in the scheme."""
        return len(self.items)

    def __getitem__(self, id_: str) -> Optional[Agency]:
        """Return the agency identified by the supplied ID."""
        out = list(filter(lambda p: p.id == id_, self.items))
        if len(out) == 0:
            return None
        else:
            return out[0]

    def __contains__(self, id_: str) -> bool:
        """Whether an agency with the supplied ID is present in the scheme."""
        return bool(self.__getitem__(id_))


class DataProviderScheme(ItemScheme, frozen=True, omit_defaults=True):
    """An immutable collection of data providers."""

    id: str = "DATA_PROVIDERS"
    name: str = "DATA_PROVIDERS"
    version: str = "1.0"
    items: Sequence[DataProvider] = ()

    @property
    def providers(self) -> Sequence[DataProvider]:
        """Extract the items in the scheme."""
        return self.items

    def __iter__(self) -> Iterator[DataProvider]:
        """Return an iterator over the list of data providers."""
        yield from self.items

    def __len__(self) -> int:
        """Return the number of data providers in the scheme."""
        return len(self.items)

    def __getitem__(self, id_: str) -> Optional[DataProvider]:
        """Return the data provider identified by the supplied ID."""
        out = list(filter(lambda p: p.id == id_, self.items))
        if len(out) == 0:
            return None
        else:
            return out[0]

    def __contains__(self, id_: str) -> bool:
        """Whether a provider with the supplied ID is present in the scheme."""
        return bool(self.__getitem__(id_))


class MetadataProviderScheme(ItemScheme, frozen=True, omit_defaults=True):
    """An immutable collection of metadata providers."""

    id: str = "METADATA_PROVIDERS"
    name: str = "METADATA_PROVIDERS"
    version: str = "1.0"
    items: Sequence[MetadataProvider] = ()

    @property
    def providers(self) -> Sequence[MetadataProvider]:
        """Extract the items in the scheme."""
        return self.items

    def __iter__(self) -> Iterator[MetadataProvider]:
        """Return an iterator over the list of metadata providers."""
        yield from self.items

    def __len__(self) -> int:
        """Return the number of metadata providers in the scheme."""
        return len(self.items)

    def __getitem__(self, id_: str) -> Optional[MetadataProvider]:
        """Return the metadata provider identified by the supplied ID."""
        out = list(filter(lambda p: p.id == id_, self.items))
        if len(out) == 0:
            return None
        else:
            return out[0]

    def __contains__(self, id_: str) -> bool:
        """Whether a provider with the supplied ID is present in the scheme."""
        return bool(self.__getitem__(id_))


class DataConsumerScheme(ItemScheme, frozen=True, omit_defaults=True):
    """An immutable collection of data consumers."""

    id: str = "DATA_CONSUMERS"
    name: str = "DATA_CONSUMERS"
    version: str = "1.0"
    items: Sequence[DataConsumer] = ()

    @property
    def consumers(self) -> Sequence[DataConsumer]:
        """Extract the items in the scheme."""
        return self.items

    def __iter__(self) -> Iterator[DataConsumer]:
        """Return an iterator over the list of data consumers."""
        yield from self.items

    def __len__(self) -> int:
        """Return the number of data consumers in the scheme."""
        return len(self.items)

    def __getitem__(self, id_: str) -> Optional[DataConsumer]:
        """Return the data consumer identified by the supplied ID."""
        out = list(filter(lambda p: p.id == id_, self.items))
        if len(out) == 0:
            return None
        else:
            return out[0]

    def __contains__(self, id_: str) -> bool:
        """Whether a consumer with the supplied ID is present in the scheme."""
        return bool(self.__getitem__(id_))



"""Module for writing metadata to XML files."""

from collections import OrderedDict
from typing import Any, Dict, Optional, Sequence, Union

from pysdmx.io.format import Format
from pysdmx.io.xml.sdmx21.__tokens import (
    AGENCIES,
    AGENCY_ID,
    AS_STATUS,
    ATT,
    ATT_REL,
    CL,
    CL_LOW,
    CLASS,
    CON,
    CON_ID,
    CONDITIONAL,
    CORE_REP,
    CS,
    DEPARTMENT,
    DFW,
    DIM,
    DSD,
    DSD_COMPS,
    EMAIL,
    ENUM,
    ENUM_FORMAT,
    FAX,
    ID,
    LOCAL_REP,
    MANDATORY,
    MEASURE,
    NAME,
    PACKAGE,
    PAR_ID,
    PAR_VER,
    POSITION,
    PRIM_MEASURE,
    REF,
    ROLE,
    RULE,
    RULE_SCHEME,
    TELEPHONE,
    TEXT_FORMAT,
    TEXT_TYPE,
    TIME_DIM,
    TRANS_SCHEME,
    TRANSFORMATION,
    UDO,
    UDO_SCHEME,
    URI,
    URN,
    VERSION,
)
from pysdmx.io.xml.sdmx21.writer.__write_aux import (
    ABBR_COM,
    ABBR_MSG,
    ABBR_STR,
    MSG_CONTENT_PKG,
    __escape_xml,
    __to_lower_camel_case,
    __write_header,
    add_indent,
    create_namespaces,
    get_end_message,
)
from pysdmx.model import (
    AgencyScheme,
    Codelist,
    Concept,
    ConceptScheme,
    DataType,
    Facets,
    Hierarchy,
    Ruleset,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
    VtlScheme,
)
from pysdmx.model.__base import (
    Agency,
    AnnotableArtefact,
    Contact,
    IdentifiableArtefact,
    Item,
    ItemReference,
    ItemScheme,
    MaintainableArtefact,
    NameableArtefact,
    Reference,
    VersionableArtefact,
)
from pysdmx.model.dataflow import (
    Component,
    Dataflow,
    DataStructureDefinition,
    Role,
)
from pysdmx.model.message import Header
from pysdmx.util import (
    parse_item_urn,
    parse_short_urn,
)

ANNOTATION_WRITER = OrderedDict(
    {
        "title": "AnnotationTitle",
        "type": "AnnotationType",
        "text": "AnnotationText",
        "url": "AnnotationURL",
    }
)

ROLE_MAPPING = {
    Role.DIMENSION: DIM,
    Role.ATTRIBUTE: ATT,
    Role.MEASURE: PRIM_MEASURE,
}

STR_TYPES = Union[
    ItemScheme,
    Codelist,
    ConceptScheme,
    DataStructureDefinition,
    Dataflow,
    RulesetScheme,
    UserDefinedOperatorScheme,
    TransformationScheme,
]

STR_DICT_TYPE_LIST = {
    AgencyScheme: "OrganisationSchemes",
    Codelist: "Codelists",
    ConceptScheme: "Concepts",
    DataStructureDefinition: "DataStructures",
    Dataflow: "Dataflows",
    RulesetScheme: "Rulesets",
    UserDefinedOperatorScheme: "UserDefinedOperators",
    TransformationScheme: "Transformations",
}


def __write_annotable(annotable: AnnotableArtefact, indent: str) -> str:
    """Writes the annotations to the XML file."""
    if len(annotable.annotations) == 0:
        return ""

    child1 = indent
    child2 = add_indent(child1)
    child3 = add_indent(child2)

    outfile = f"{child1}<{ABBR_COM}:Annotations>"
    for annotation in annotable.annotations:
        if annotation.id is None:
            outfile += f"{child2}<{ABBR_COM}:Annotation>"
        else:
            outfile += (
                f"{child2}<{ABBR_COM}:Annotation " f"id={annotation.id!r}>"
            )
        outfile = outfile.replace("'", '"')

        for attr, label in ANNOTATION_WRITER.items():
            if getattr(annotation, attr, None) is not None:
                value = getattr(annotation, attr)
                value = __escape_xml(str(value))

                if attr == "text":
                    head_tag = f'{ABBR_COM}:{label} xml:lang="en"'

                else:
                    head_tag = f"{ABBR_COM}:{label}"

                outfile += (
                    f"{child3}<{head_tag}>" f"{value}" f"</{ABBR_COM}:{label}>"
                )

        outfile += f"{child2}</{ABBR_COM}:Annotation>"
    outfile += f"{child1}</{ABBR_COM}:Annotations>"
    return outfile


def __write_identifiable(
    identifiable: IdentifiableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the IdentifiableArtefact to the XML file."""
    attributes = ""

    attributes += f" id={identifiable.id!r}"

    if identifiable.uri is not None:
        attributes += f" uri={identifiable.uri!r}"

    if identifiable.urn is not None:
        attributes += f" urn={identifiable.urn!r}"

    outfile = {
        "Annotations": __write_annotable(identifiable, indent),
        "Attributes": attributes,
    }

    return outfile


def __write_nameable(
    nameable: NameableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the NameableArtefact to the XML file."""
    outfile = __write_identifiable(nameable, indent)
    attrs = ["Name", "Description"]

    for attr in attrs:
        if getattr(nameable, attr.lower(), None) is not None:
            value = getattr(nameable, attr.lower())
            value = __escape_xml(str(value))
            outfile[attr] = [
                (
                    f"{indent}"
                    f'<{ABBR_COM}:{attr} xml:lang="en">'
                    f"{value}"
                    f"</{ABBR_COM}:{attr}>"
                )
            ]

    return outfile


def __write_versionable(
    versionable: VersionableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the VersionableArtefact to the XML file."""
    outfile = __write_nameable(versionable, add_indent(indent))

    outfile["Attributes"] += f" version={versionable.version!r}"

    if versionable.valid_from is not None:
        valid_from_str = versionable.valid_from.strftime("%Y-%m-%dT%H:%M:%S")
        outfile["Attributes"] += f" validFrom={valid_from_str!r}"

    if versionable.valid_to is not None:
        valid_to_str = versionable.valid_to.strftime("%Y-%m-%dT%H:%M:%S")
        outfile["Attributes"] += f" validTo={valid_to_str!r}"

    return outfile


def __write_maintainable(
    maintainable: MaintainableArtefact, indent: str
) -> Dict[str, Any]:
    """Writes the MaintainableArtefact to the XML file."""
    outfile = __write_versionable(maintainable, indent)

    outfile["Attributes"] += (
        f" isExternalReference="
        f"{str(maintainable.is_external_reference).lower()!r}"
    )

    outfile["Attributes"] += f" isFinal={str(maintainable.is_final).lower()!r}"

    if isinstance(maintainable.agency, str):
        outfile["Attributes"] += f" agencyID={maintainable.agency!r}"
    else:
        outfile["Attributes"] += f" agencyID={maintainable.agency.id!r}"

    return outfile


def __write_contact(contact: Contact, indent: str) -> str:
    """Writes the contact to the XML file."""

    def __item_to_str(item: Any, ns: str, tag: str) -> str:
        return (
            f"{add_indent(indent)}"
            f"<{ns}:{tag}>"
            f"{__escape_xml(str(item))}"
            f"</{ns}:{tag}>"
        )

    def __items_to_str(items: Any, ns: str, tag: str) -> str:
        return "".join([__item_to_str(item, ns, tag) for item in items])

    outfile = f"{indent}<{ABBR_STR}:Contact>"
    if contact.name is not None:
        outfile += __item_to_str(contact.name, ABBR_COM, NAME)
    if contact.department is not None:
        outfile += __item_to_str(contact.department, ABBR_STR, DEPARTMENT)
    if contact.role is not None:
        outfile += __item_to_str(contact.role, ABBR_STR, ROLE)
    if contact.telephones is not None:
        outfile += __items_to_str(contact.telephones, ABBR_STR, TELEPHONE)
    if contact.faxes is not None:
        outfile += __items_to_str(contact.faxes, ABBR_STR, FAX)
    if contact.uris is not None:
        outfile += __items_to_str(contact.uris, ABBR_STR, URI)
    if contact.emails is not None:
        outfile += __items_to_str(contact.emails, ABBR_STR, EMAIL)
    outfile += f"{indent}</{ABBR_STR}:Contact>"

    return outfile


def __write_item(item: Item, indent: str) -> str:
    """Writes the item to the XML file."""
    head = f"{ABBR_STR}:" + type(item).__name__

    data = __write_nameable(item, add_indent(indent))
    attributes = data["Attributes"].replace("'", '"')
    outfile = f"{indent}<{head}{attributes}>"
    outfile += __export_intern_data(data)
    if isinstance(item, Agency) and len(item.contacts) > 0:
        for contact in item.contacts:
            outfile += __write_contact(contact, add_indent(indent))
    if isinstance(item, Concept) and (
        item.codes is not None
        or item.facets is not None
        or item.dtype is not None
    ):
        outfile += f"{add_indent(indent)}<{ABBR_STR}:{CORE_REP}>"
        if item.codes is not None:
            outfile += __write_enumeration(item.codes, add_indent(indent))
        if item.facets is not None or item.dtype is not None:
            outfile += __write_text_format(
                item.dtype, item.facets, TEXT_FORMAT, add_indent(indent)
            )
        outfile += f"{add_indent(indent)}</{ABBR_STR}:{CORE_REP}>"
    outfile += f"{indent}</{head}>"
    return outfile


def __write_components(item: DataStructureDefinition, indent: str) -> str:
    """Writes the components to the XML file."""
    outfile = f"{indent}<{ABBR_STR}:{DSD_COMPS}>"

    components: Dict[str, Any] = {
        DIM: [],
        ATT: [],
        PRIM_MEASURE: [],
    }

    for comp in item.components:
        if comp.role == Role.DIMENSION:
            components[DIM].append(comp)
        elif comp.role == Role.ATTRIBUTE:
            components[ATT].append(comp)
        else:
            components[PRIM_MEASURE].append(comp)

    position = 1
    for _, comps in components.items():
        if comps:
            role_name = ROLE_MAPPING[comps[0].role]
            if role_name == PRIM_MEASURE:
                role_name = MEASURE
            outfile += f"{add_indent(indent)}<{ABBR_STR}:{role_name}List>"
            for comp in comps:
                outfile += __write_component(
                    comp, position, add_indent(add_indent(indent)), components
                )
                position += 1
            outfile += f"{add_indent(indent)}</{ABBR_STR}:{role_name}List>"

    outfile += f"{indent}</{ABBR_STR}:{DSD_COMPS}>"
    return outfile


def __write_attribute_relation(
    item: Component, indent: str, component_info: Dict[str, Any]
) -> str:
    outfile = f"{indent}<{ABBR_STR}:{ATT_REL}>"
    att_rel = item.attachment_level
    if att_rel is None or att_rel == "D":
        outfile += f"{add_indent(indent)}<{ABBR_STR}:None/>"
    else:
        # Check if it is a list of Dimensions or it is related to the
        # primary measure
        if "," in att_rel:
            comps_to_relate = att_rel.split(",")
        elif att_rel == "O":
            comps_to_relate = [component_info[PRIM_MEASURE][0].id]
        else:
            comps_to_relate = [att_rel]

        dim_names = [comp.id for comp in component_info[DIM]]

        for comp_name in comps_to_relate:
            role = Role.DIMENSION if comp_name in dim_names else Role.MEASURE
            related_role = ROLE_MAPPING[role]
            outfile += f"{add_indent(indent)}<{ABBR_STR}:{related_role}>"
            outfile += (
                f"{add_indent(add_indent(indent))}"
                f"<{REF} {ID}={comp_name!r}/>"
            )
            outfile += f"{add_indent(indent)}</{ABBR_STR}:{related_role}>"
    outfile += f"{indent}</{ABBR_STR}:{ATT_REL}>"

    return outfile


def __write_component(
    item: Component, position: int, indent: str, component_info: Dict[str, Any]
) -> str:
    """Writes the component to the XML file."""
    role_name = ROLE_MAPPING[item.role]
    if role_name == DIM and item.id == "TIME_PERIOD":
        role_name = TIME_DIM
    head = f"{indent}<{ABBR_STR}:{role_name} "

    attributes = ""
    attribute_relation = ""
    if item.role == Role.ATTRIBUTE:
        status = MANDATORY if item.required else CONDITIONAL
        attributes += f"{AS_STATUS}={status!r} "
        attribute_relation = __write_attribute_relation(
            item, add_indent(indent), component_info
        )

    attributes += f"{ID}={item.id!r}"
    if item.role == Role.DIMENSION:
        attributes += f" {POSITION}={str(position)!r}"

    if item.urn is not None:
        attributes += f" {URN.lower()}={item.urn!r}"

    attributes += ">"

    concept_identity = __write_concept_identity(
        item.concept, add_indent(indent)
    )
    representation = __write_representation(item, add_indent(indent))

    outfile = head
    outfile += attributes
    outfile += concept_identity
    outfile += representation
    outfile += attribute_relation

    outfile += f"{indent}</{ABBR_STR}:{role_name}>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_concept_identity(
    identity: Union[Concept, ItemReference], indent: str
) -> str:
    if isinstance(identity, ItemReference):
        ref = identity
    else:
        ref = parse_item_urn(identity.urn)  # type: ignore[arg-type]

    outfile = f"{indent}<{ABBR_STR}:{CON_ID}>"
    outfile += f"{add_indent(indent)}<{REF} "
    outfile += f"{AGENCY_ID}={ref.agency!r} "
    outfile += f"{CLASS}={CON!r} "
    outfile += f"{ID}={ref.item_id!r} "
    outfile += f"{PAR_ID}={ref.id!r} "
    outfile += f"{PAR_VER}={ref.version!r} "
    outfile += f"{PACKAGE}={CS.lower()!r}/>"
    outfile += f"{indent}</{ABBR_STR}:{CON_ID}>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_representation(item: Component, indent: str) -> str:
    representation = ""
    local_representation = ""

    if item.local_codes is not None:
        local_representation += __write_enumeration(item.local_codes, indent)

    if item.local_facets is not None or item.local_dtype is not None:
        type_ = ENUM_FORMAT if item.local_codes is not None else TEXT_FORMAT
        local_representation += __write_text_format(
            item.local_dtype, item.local_facets, type_, indent
        )

    representation += f"{indent}<{ABBR_STR}:{LOCAL_REP}>"
    if len(local_representation) == 0:
        representation += f"{add_indent(indent)}<{ABBR_STR}:{TEXT_FORMAT}/>"
    representation += local_representation
    representation += f"{indent}</{ABBR_STR}:{LOCAL_REP}>"

    return representation


def __write_text_format(
    dtype: Optional[DataType],
    facets: Optional[Facets],
    type_: str,
    indent: str,
) -> str:
    """Writes the text format to the XML file."""
    outfile = f"{add_indent(indent)}<{ABBR_STR}:{type_}"
    if facets is not None:
        active_facets = facets.__str__().replace("=", '="').split(", ")
        for facet in active_facets:
            facet = __to_lower_camel_case(facet)
            outfile += f' {facet}"'
    if dtype is not None:
        outfile += f" {TEXT_TYPE}={dtype.value!r}"
    outfile += "/>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_enumeration(codes: Union[Codelist, Hierarchy], indent: str) -> str:
    """Writes the enumeration to the XML file."""
    ref = parse_short_urn(codes.short_urn)

    outfile = f"{add_indent(indent)}<{ABBR_STR}:{ENUM}>"
    outfile += f"{add_indent(add_indent(indent))}<{REF} "
    outfile += f"{AGENCY_ID}={ref.agency!r} "
    outfile += f"{CLASS}={CL!r} "
    outfile += f"{ID}={ref.id!r} "
    outfile += f"{PACKAGE}={CL_LOW!r} "
    outfile += f"{VERSION}={ref.version!r}/>"
    outfile += f"{add_indent(indent)}</{ABBR_STR}:{ENUM}>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_structure(item: str, indent: str) -> str:
    """Writes the dataflow structure to the XML file."""
    ref = parse_short_urn(item)
    outfile = f"{indent}<{ABBR_STR}:Structure>"
    outfile += (
        f"{add_indent(indent)}<{REF} "
        f'{PACKAGE}="datastructure" '
        f"{AGENCY_ID}={ref.agency!r} "
        f"{ID}={ref.id!r} "
        f"{VERSION}={ref.version!r} "
        f"{CLASS}={DSD!r}/>"
    )
    outfile += f"{indent}</{ABBR_STR}:Structure>"

    outfile = outfile.replace("'", '"')
    return outfile


def __write_scheme(item_scheme: Any, indent: str, scheme: str) -> str:
    """Writes the scheme to the XML file."""
    label = f"{ABBR_STR}:{scheme}"
    components = ""
    data = __write_maintainable(item_scheme, indent)

    if scheme == DSD:
        components = __write_components(item_scheme, add_indent(indent))

    if scheme not in [DSD, DFW, RULE_SCHEME, UDO_SCHEME, TRANS_SCHEME]:
        data["Attributes"] += (
            f" isPartial={str(item_scheme.is_final).lower()!r}"
        )
    if scheme in [RULE_SCHEME, UDO_SCHEME, TRANS_SCHEME]:
        data["Attributes"] += f" {_write_vtl(item_scheme, indent)}"

    outfile = ""

    attributes = data.get("Attributes") or ""
    attributes = attributes.replace("'", '"')

    outfile += f"{indent}<{label}{attributes}>"

    outfile += __export_intern_data(data)

    outfile += components

    if scheme == DFW:
        outfile += __write_structure(item_scheme.structure, add_indent(indent))

    if scheme not in [DSD, DFW, RULE_SCHEME, UDO_SCHEME, TRANS_SCHEME]:
        for item in item_scheme.items:
            outfile += __write_item(item, add_indent(indent))
    if scheme in [RULE_SCHEME, UDO_SCHEME, TRANS_SCHEME]:
        for item in item_scheme.items:
            outfile += _write_vtl(item, add_indent(indent))
        outfile += _write_vtl_references(item_scheme, add_indent(indent))
    outfile += f"{indent}</{label}>"

    return outfile


def __write_metadata_element(
    package: Dict[str, Any], key: str, prettyprint: object
) -> str:
    """Writes the metadata element to the XML file.

    Args:
        package: The package to be written
        key: The key to be used
        prettyprint: Prettyprint or not

    Returns:
        A string with the metadata element
    """
    outfile = ""
    nl = "\n" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""

    base_indent = f"{nl}{child2}"

    if key in package:
        outfile += f"{base_indent}<{ABBR_STR}:{MSG_CONTENT_PKG[key]}>"
        for element in package[key].values():
            item = (
                DSD
                if issubclass(element.__class__, DataStructureDefinition)
                else (
                    AGENCIES
                    if element.id == "AGENCIES"
                    else type(element).__name__
                )
            )
            outfile += __write_scheme(element, add_indent(base_indent), item)
        outfile += f"{base_indent}</{ABBR_STR}:{MSG_CONTENT_PKG[key]}>"

    return outfile


def __get_outfile(obj_: Dict[str, Any], key: str = "") -> str:
    """Generates an outfile from the object.

    Args:
        obj_: The object to be used
        key: The key to be used

    Returns:
        A string with the outfile

    """
    element = obj_.get(key) or []

    return "".join(element)


def __export_intern_data(data: Dict[str, Any]) -> str:
    """Export internal data (Annotations, Name, Description) on the XML file.

    Args:
        data: Information to be exported

    Returns:
        The XML string with the exported data
    """
    outfile = __get_outfile(data, "Annotations")
    outfile += __get_outfile(data, "Name")
    outfile += __get_outfile(data, "Description")

    return outfile


def __write_structures(content: Dict[str, Any], prettyprint: bool) -> str:
    """Writes the structures to the XML file.

    Args:
        content: The Message Content to be written
        prettyprint: Prettyprint or not

    Returns:
        A string with the structures
    """
    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""

    outfile = f"{nl}{child1}<{ABBR_MSG}:Structures>"

    for key in MSG_CONTENT_PKG:
        outfile += __write_metadata_element(content, key, prettyprint)

    outfile += f"{nl}{child1}</{ABBR_MSG}:Structures>"

    # Replace &amp; with & in the outfile
    outfile = outfile.replace("& ", "&amp; ")

    return outfile


def _write_vtl(item_or_scheme: Union[Item, ItemScheme], indent: str) -> str:
    """Writes the VTL attribute to the XML file for a single item.

    This function writes an item or an item scheme to the XML file,
    following the standard format.

    Args:
        item_or_scheme: The item or item scheme to be written
            Item: The item to be written
            ItemScheme: The item scheme to be written
        indent: The current indentation level
    """
    outfile = ""

    if isinstance(item_or_scheme, Item):
        label = ""
        nameable = __write_nameable(item_or_scheme, add_indent(indent))
        attrib = nameable["Attributes"].replace("'", '"')
        data = __export_intern_data(nameable)

        if isinstance(item_or_scheme, Ruleset):
            label = f"{ABBR_STR}:{RULE}"
            data += f"{add_indent(indent)}<{ABBR_STR}:RulesetDefinition>"
            data += (
                f"{__escape_xml(item_or_scheme.ruleset_definition)}"
                f"</{ABBR_STR}:RulesetDefinition>"
            )
            attrib += (
                f" rulesetScope={item_or_scheme.ruleset_scope!r} "
                f"rulesetType={item_or_scheme.ruleset_type!r}"
            )

        if isinstance(item_or_scheme, Transformation):
            label = f"{ABBR_STR}:{TRANSFORMATION}"
            data += f"{add_indent(indent)}<{ABBR_STR}:Expression>"
            data += (
                f"{__escape_xml(item_or_scheme.expression)}"
                f"</{ABBR_STR}:Expression>"
            )
            data += f"{add_indent(indent)}<{ABBR_STR}:Result>"
            data += f"{item_or_scheme.result}</{ABBR_STR}:Result>"
            attrib += f" isPersistent={item_or_scheme.is_persistent!r}"

        if isinstance(item_or_scheme, UserDefinedOperator):
            label = f"{ABBR_STR}:{UDO}"
            data += f"{add_indent(indent)}<{ABBR_STR}:OperatorDefinition>"
            data += (
                f"{__escape_xml(item_or_scheme.operator_definition)}"
                f"</{ABBR_STR}:OperatorDefinition>"
            )

        outfile += f"{indent}<{label}{attrib}>"
        outfile += data
        outfile += f"{indent}</{label}>"

    if isinstance(item_or_scheme, VtlScheme):
        outfile += f" vtlVersion={item_or_scheme.vtl_version!r}"

    outfile = outfile.replace("'", '"')

    return outfile


def _write_vtl_references(scheme: ItemScheme, indent: str) -> str:
    """Writes references to VTL elements to the XML file."""

    def process_references(
        references: Union[Any, Sequence[Any]], element_name: str
    ) -> str:
        """Process the references to VTL elements."""
        outreference = []
        if not isinstance(references, (list, tuple)):
            references = [references]

        for ref in references:
            if isinstance(ref, Reference):
                outreference.append(
                    f"{indent}<{ABBR_STR}:{element_name}>"
                    f"{add_indent(indent)}<{REF} "
                    f"{PACKAGE}={TRANSFORMATION.lower()!r} "
                    f"{AGENCY_ID}={ref.agency!r} "
                    f"{ID}={ref.id!r} "
                    f"{VERSION}={ref.version!r} "
                    f"{CLASS}={ref.sdmx_type!r}/>"
                    f"{indent}</{ABBR_STR}:{element_name}>"
                )
            if isinstance(ref, ItemScheme):
                ref_to_use = parse_short_urn(ref.short_urn)
                outreference.append(
                    f"{indent}<{ABBR_STR}:{element_name}>"
                    f"{add_indent(indent)}<{REF} "
                    f"{PACKAGE}={TRANSFORMATION.lower()!r} "
                    f"{AGENCY_ID}={ref_to_use.agency!r} "
                    f"{ID}={ref_to_use.id!r} "
                    f"{VERSION}={ref_to_use.version!r} "
                    f"{CLASS}={ref_to_use.sdmx_type!r}/>"
                    f"{indent}</{ABBR_STR}:{element_name}>"
                )

        return "".join(outreference)

    outfile = ""
    if isinstance(scheme, TransformationScheme):
        outfile += process_references(scheme.ruleset_schemes, "RulesetScheme")
        outfile += process_references(
            scheme.user_defined_operator_schemes, "UserDefinedOperatorScheme"
        )
    if isinstance(scheme, UserDefinedOperatorScheme):
        outfile += process_references(scheme.ruleset_schemes, "RulesetScheme")

    outfile = outfile.replace("'", '"')

    return outfile


def write(
    structures: Sequence[STR_TYPES],
    output_path: str = "",
    prettyprint: bool = True,
    header: Optional[Header] = None,
) -> Optional[str]:
    """This function writes a SDMX-ML file from the Message Content.

    Args:
        structures: The content to be written
        output_path: The path to save the file
        prettyprint: Prettyprint or not
        header: The header to be used (generated if None)

    Returns:
        The XML string if output_path is empty, None otherwise
    """
    type_ = Format.STRUCTURE_SDMX_ML_2_1
    elements = {structure.short_urn: structure for structure in structures}
    if header is None:
        header = Header()

    content: Dict[str, Dict[str, STR_TYPES]] = {}
    for urn, element in elements.items():
        list_ = STR_DICT_TYPE_LIST[type(element)]
        if list_ not in content:
            content[list_] = {}
        content[list_][urn] = element

    # Generating the initial tag with namespaces
    outfile = create_namespaces(type_, prettyprint=prettyprint)
    # Generating the header
    outfile += __write_header(header, prettyprint, data_message=False)
    # Writing the content
    outfile += __write_structures(content, prettyprint)

    outfile += get_end_message(type_, prettyprint)

    if output_path == "":
        return outfile
    with open(output_path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)
    return None



# mypy: disable-error-code="union-attr"
"""Writer auxiliary functions."""

import warnings
from collections import OrderedDict
from datetime import datetime
from typing import Optional, Union
from xml.sax.saxutils import escape

from pysdmx.errors import Invalid, NotImplemented
from pysdmx.io.format import Format
from pysdmx.io.xml.sdmx21.__tokens import (
    ANNOTATIONS_LOW,
    CONTACTS_LOW,
    DESC_LOW,
    DFW,
    DFWS_LOW,
    DSD,
    PROV_AGREMENT,
    RULESETS,
    STR_USAGE,
    STRUCTURE,
    TRANSFORMATIONS,
    UDOS,
    URI_LOW,
    URN_LOW,
)
from pysdmx.model import Organisation
from pysdmx.model.dataset import Dataset
from pysdmx.model.message import Header
from pysdmx.util import parse_short_urn

MESSAGE_TYPE_MAPPING = {
    Format.DATA_SDMX_ML_2_1_GEN: "GenericData",
    Format.DATA_SDMX_ML_2_1_STR: "StructureSpecificData",
    Format.STRUCTURE_SDMX_ML_2_1: "Structure",
    Format.ERROR_SDMX_ML_2_1: "Error",
    Format.REGISTRY_SDMX_ML_2_1: "RegistryInterface",
}

ABBR_MSG = "mes"
ABBR_GEN = "gen"
ABBR_COM = "com"
ABBR_STR = "str"
ABBR_SPE = "ss"

ANNOTATIONS = "Annotations"
STRUCTURES = "Structures"
ORGS = "OrganisationSchemes"
AGENCIES = "AgencyScheme"
CODELISTS = "Codelists"
CONCEPTS = "Concepts"
DSDS = "DataStructures"
DATAFLOWS = "Dataflows"
CONSTRAINTS = "Constraints"
ALL_DIM = "AllDimensions"

BASE_URL = "http://www.sdmx.org/resources/sdmxml/schemas/v2_1"

NAMESPACES = {
    "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    ABBR_MSG: f"{BASE_URL}/message",
    ABBR_GEN: f"{BASE_URL}/data/generic",
    ABBR_COM: f"{BASE_URL}/common",
    ABBR_STR: f"{BASE_URL}/structure",
    ABBR_SPE: f"{BASE_URL}/data/structurespecific",
}

URN_DS_BASE = "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="


def __namespaces_from_type(type_: Format) -> str:
    """Returns the namespaces for the XML file based on type.

    Args:
        type_: MessageType to be used

    Returns:
        A string with the namespaces

    Raises:
        NotImplemented: If the MessageType is not implemented
    """
    if type_ == Format.STRUCTURE_SDMX_ML_2_1:
        return f"xmlns:{ABBR_STR}={NAMESPACES[ABBR_STR]!r} "
    elif type_ == Format.DATA_SDMX_ML_2_1_STR:
        return f"xmlns:{ABBR_SPE}={NAMESPACES[ABBR_SPE]!r} "
    elif type_ == Format.DATA_SDMX_ML_2_1_GEN:
        return f"xmlns:{ABBR_GEN}={NAMESPACES[ABBR_GEN]!r} "
    else:
        raise NotImplemented(f"{type_} not implemented")


def create_namespaces(
    type_: Format, ss_namespaces: str = "", prettyprint: bool = False
) -> str:
    """Creates the namespaces for the XML file.

    Args:
        type_: MessageType to be used
        ss_namespaces: The namespaces for the StructureSpecificData
        prettyprint: Prettyprint or not

    Returns:
        A string with the namespaces
    """
    nl = "\n" if prettyprint else ""

    outfile = f'<?xml version="1.0" encoding="UTF-8"?>{nl}'

    outfile += f"<{ABBR_MSG}:{MESSAGE_TYPE_MAPPING[type_]} "
    outfile += f'xmlns:xsi={NAMESPACES["xsi"]!r} '
    outfile += f"xmlns:{ABBR_MSG}={NAMESPACES[ABBR_MSG]!r} "
    outfile += __namespaces_from_type(type_)
    outfile += (
        f"xmlns:{ABBR_COM}={NAMESPACES[ABBR_COM]!r} "
        f"{ss_namespaces}"
        f'xsi:schemaLocation="{NAMESPACES[ABBR_MSG]} '
        f'https://registry.sdmx.org/schemas/v2_1/SDMXMessage.xsd">'
    )

    return outfile.replace("'", '"')


MSG_CONTENT_PKG = OrderedDict(
    [
        (ORGS, "OrganisationSchemes"),
        (DATAFLOWS, "Dataflows"),
        (CODELISTS, "Codelists"),
        (CONCEPTS, "Concepts"),
        (DSDS, "DataStructures"),
        (CONSTRAINTS, "ContentConstraints"),
        (RULESETS, "Rulesets"),
        (TRANSFORMATIONS, "Transformations"),
        (UDOS, "UserDefinedOperators"),
    ]
)


def get_end_message(type_: Format, prettyprint: bool) -> str:
    """Returns the end message for the XML file.

    Args:
        type_: MessageType to be used
        prettyprint: Prettyprint or not

    Returns:
        A string with the end message
    """
    nl = "\n" if prettyprint else ""
    return f"{nl}</{ABBR_MSG}:{MESSAGE_TYPE_MAPPING[type_]}>"


def add_indent(indent: str) -> str:
    """Adds another indent.

    Args:
        indent: The string to be indented

    Returns:
        A string with one more indentation
    """
    return indent + "\t"


def __item(
    element: str,
    sender_receiver: Optional[Union[str, Organisation]],
    nl: str,
    prettyprint: bool,
) -> str:
    """Generates an item element for the XML file.

    An Item element is an XML tag with an id attribute.

    Args:
        element: Sender, Receiver
        sender_receiver: The sender or receiver to be written
        nl: Newline character
        prettyprint: Prettyprint or not

    Returns:
        A string with the item element
    """
    child2 = "\t\t" if prettyprint else ""
    child3 = "\t\t\t" if prettyprint else ""
    name = None
    org_id = None

    if isinstance(sender_receiver, Organisation):
        org_id = sender_receiver.id
        name = sender_receiver.name

        unexpected_keys = {
            URI_LOW,
            URN_LOW,
            DESC_LOW,
            CONTACTS_LOW,
            DFWS_LOW,
            ANNOTATIONS_LOW,
        }

        unexpected_with_values = {
            key
            for key in unexpected_keys
            if hasattr(sender_receiver, key)
            and getattr(sender_receiver, key) not in (None, (), [], {})
        }

        if unexpected_with_values:
            warnings.warn(
                f"The following attributes will be lost: "
                f"{', '.join(unexpected_with_values)}",
                UserWarning,
                stacklevel=2,
            )
    else:
        return ""

    message = f"{nl}{child2}<{ABBR_MSG}:{element} id={org_id!r}"
    if name is not None:
        message += (
            f">"
            f"{nl}{child3}<{ABBR_COM}:Name>{name}</{ABBR_COM}:Name>"
            f"{nl}{child2}</{ABBR_MSG}:{element}>"
        )
    else:
        message += "/>"
    return message


def __reference(
    urn_structure: str,
    dimension: Optional[str],
    nl: str,
    prettyprint: bool,
    add_namespace_structure: bool,
) -> str:
    child2 = "\t\t" if prettyprint else ""
    child3 = "\t\t\t" if prettyprint else ""
    child4 = "\t\t\t\t" if prettyprint else ""
    namespace = ""
    reference = parse_short_urn(urn_structure)
    if reference.sdmx_type == DSD:
        structure_type = STRUCTURE
    elif reference.sdmx_type == DFW:
        structure_type = STR_USAGE
    else:
        structure_type = PROV_AGREMENT
    if add_namespace_structure:
        namespace = (
            f"{URN_DS_BASE}={reference.agency}:{reference.id}"
            f"({reference.version})"
        )
        namespace = f"namespace={namespace!r} "

    return (
        # First the message structure
        f"{nl}{child2}<{ABBR_MSG}:Structure "
        f"structureID={reference.id!r} "
        f"{namespace}"
        f"dimensionAtObservation={dimension!r}>"
        # Then the common structure
        f"{nl}{child3}<{ABBR_COM}:{structure_type}>"
        # Then the reference
        f"{nl}{child4}<Ref agencyID={reference.agency!r} "
        f"id={reference.id!r} version={reference.version!r} "
        f"class={reference.sdmx_type!r}/>"
        # Close the common structure
        f"{nl}{child3}</{ABBR_COM}:{structure_type}>"
        # Close the message structure
        f"{nl}{child2}</{ABBR_MSG}:Structure>"
    )


def __write_header(
    header: Header,
    prettyprint: bool,
    add_namespace_structure: bool = False,
    data_message: bool = True,
) -> str:
    """Writes the Header part of the message.

    Args:
        header: The Header to be written
        prettyprint: Prettyprint or not
        add_namespace_structure: Add the namespace for the structure
        data_message: If the message is a data message

    Returns:
        The XML string
    """

    def __value(element: str, value: Optional[str]) -> str:
        """Generates a value element for the XML file.

        A Value element is an XML tag with a value.

        Args:
            element: ID, Test, Prepared, Sender, Receiver, Source
            value: The value to be written

        Returns:
            A string with the value element
        """
        if not value:
            return ""
        child2 = "\t\t" if prettyprint else ""
        return (
            f"{nl}{child2}<{ABBR_MSG}:{element}>"
            f"{value}"
            f"</{ABBR_MSG}:{element}>"
        )

    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""
    if isinstance(header.prepared, str):
        prepared = datetime.strptime(header.prepared, "%Y-%m-%dT%H:%M:%S")
    else:
        prepared = header.prepared.strftime("%Y-%m-%dT%H:%M:%S")
    test = str(header.test).lower()
    references_str = ""
    action_value = (
        header.dataset_action.value if header.dataset_action else None
    )
    if header.structure is not None:
        for short_urn, dim_at_obs in header.structure.items():
            references_str += __reference(
                short_urn,
                dim_at_obs,
                nl,
                prettyprint,
                add_namespace_structure,
            )
    if not data_message and (
        header.dataset_id or header.dataset_action or header.structure
    ):
        raise Invalid(
            "Header must not contain DataSetID or DataSetAction "
            "when writing a Structures Message."
        )
    return (
        f"{nl}{child1}<{ABBR_MSG}:Header>"
        f"{__value('ID', header.id)}"
        f"{__value('Test', test)}"
        f"{__value('Prepared', prepared)}"
        f"{__item('Sender', header.sender, nl, prettyprint)}"
        f"{__item('Receiver', header.receiver, nl, prettyprint)}"
        f"{references_str}"
        f"{__value('DataSetAction', action_value)}"
        f"{__value('DataSetID', header.dataset_id)}"
        f"{__value('Source', header.source)}"
        f"{nl}{child1}</{ABBR_MSG}:Header>"
    ).replace("'", '"')


# ------------------
# -- DATA WRITING --
# ------------------


def get_structure(dataset: Dataset) -> str:
    """This function gets the structure of a dataset.

    Args:
        dataset: The dataset to get the structure from

    Returns:
        The structure Short URN
    """
    if isinstance(dataset.structure, str):
        return dataset.structure
    return dataset.structure.short_urn


def __to_camel_case(snake_str: str) -> str:
    return "".join(x.capitalize() for x in snake_str.lower().split("_"))


def __to_lower_camel_case(snake_str: str) -> str:
    # We capitalize the first letter of each component except the first one
    # with the 'capitalize' method and join them together.
    camel_string = __to_camel_case(snake_str)
    return snake_str[0].lower() + camel_string[1:]


def __escape_xml(value: str) -> str:
    return escape(value)



"""Private module for VTL validation functions."""

from vtlengine.API import create_ast  # type: ignore[import-untyped]
from vtlengine.AST import (  # type: ignore[import-untyped]
    DPRuleset as ASTDPRuleset,
)
from vtlengine.AST import HRuleset as ASTHRuleset
from vtlengine.AST import (
    Operator as ASTOperator,
)

from pysdmx.errors import Invalid
from pysdmx.io.xml.sdmx21.__tokens import RULE_SCHEME, UDO_SCHEME
from pysdmx.model import Reference
from pysdmx.model.vtl import (
    Ruleset,
    RulesetScheme,
    Transformation,
    TransformationScheme,
    UserDefinedOperator,
    UserDefinedOperatorScheme,
)


def _ruleset_validation(ruleset: Ruleset) -> None:
    """Additional validation checks for rulesets."""
    try:
        ast = create_ast(ruleset.ruleset_definition)
    except Exception as e:
        raise Invalid(f"Invalid ruleset definition: {str(e)}") from e
    if len(ast.children) > 1:
        raise Invalid("A single RulesetDefinition is valid in a Ruleset")
    if ruleset.ruleset_type == "hierarchical" and not isinstance(
        ast.children[0], ASTHRuleset
    ):
        raise Invalid("Ruleset type does not match the definition")
    if ruleset.ruleset_type == "datapoint" and not isinstance(
        ast.children[0], ASTDPRuleset
    ):
        raise Invalid("Ruleset type does not match the definition")
    if (
        ruleset.ruleset_scope == "variable"
        and ast.children[0].signature_type != "variable"
    ):
        raise Invalid("Ruleset scope does not match the definition")
    if (
        ruleset.ruleset_scope == "valuedomain"
        and ast.children[0].signature_type != "valuedomain"
    ):
        raise Invalid("Ruleset scope does not match the definition")


def _ruleset_scheme_validations(ruleset_scheme: RulesetScheme) -> None:
    """Additional validation checks for ruleset schemes."""
    if not ruleset_scheme.items:
        raise Invalid("RulesetScheme must contain at least one Ruleset")
    for ruleset in ruleset_scheme.items:
        if not isinstance(ruleset, Ruleset):
            raise Invalid("RulesetScheme must contain Ruleset items")
        _ruleset_validation(ruleset)


def _ruleset_scheme_reference_validations(ruleset_scheme: Reference) -> None:
    """Additional validation checks for ruleset schemes."""
    if ruleset_scheme.sdmx_type != RULE_SCHEME:
        raise Invalid(
            "Reference in Ruleset Schemes must point to a Ruleset Scheme, "
            f"got {ruleset_scheme.sdmx_type}"
        )


def _user_defined_operator_validation(udo: UserDefinedOperator) -> None:
    """Additional validation checks for user defined operators."""
    try:
        ast = create_ast(udo.operator_definition)
    except Exception as e:
        raise Invalid(f"Invalid operator definition: {str(e)}") from e
    if len(ast.children) > 1:
        raise Invalid(
            "A single OperatorDefinition is valid in a UserDefinedOperator"
        )
    if not isinstance(ast.children[0], ASTOperator):
        raise Invalid(
            "User defined operator type does not match the definition"
        )


def _user_defined_operator_scheme_validations(
    udo_scheme: UserDefinedOperatorScheme,
) -> None:
    """Additional validation checks for user defined operator schemes."""
    if not udo_scheme.items:
        raise Invalid(
            "UserDefinedOperatorScheme must contain at least one "
            "UserDefinedOperator"
        )
    for udo in udo_scheme.items:
        if not isinstance(udo, UserDefinedOperator):
            raise Invalid(
                "UserDefinedOperatorScheme must contain UserDefinedOperator "
                "items"
            )
        _user_defined_operator_validation(udo)


def _udo_scheme_reference_validations(udo_scheme: Reference) -> None:
    """Additional validation checks for ruleset schemes."""
    if udo_scheme.sdmx_type != UDO_SCHEME:
        raise Invalid(
            "Reference in User Defined Operator Schemes"
            " must point to a Defined Operator Scheme, "
            f"got {udo_scheme.sdmx_type}"
        )


def _transformation_validations(transformation: Transformation) -> None:
    """Additional validation checks for transformations."""
    try:
        ast = create_ast(transformation.full_expression)
    except Exception as e:
        raise Invalid(f"Invalid transformation definition: {str(e)}") from e
    if len(ast.children) > 1:
        raise Invalid("A single Assignment is valid in a Transformation")


def _transformation_scheme_validations(
    transformation_scheme: TransformationScheme,
) -> None:
    """Additional validation checks for transformation schemes."""
    if not transformation_scheme.items:
        raise Invalid(
            "TransformationScheme must contain at least one Transformation"
        )

    for ruleset_scheme in transformation_scheme.ruleset_schemes:
        if isinstance(ruleset_scheme, RulesetScheme):
            _ruleset_scheme_validations(ruleset_scheme)
        if isinstance(ruleset_scheme, Reference):
            _ruleset_scheme_reference_validations(ruleset_scheme)
    for udo_scheme in transformation_scheme.user_defined_operator_schemes:
        if isinstance(udo_scheme, UserDefinedOperatorScheme):
            _user_defined_operator_scheme_validations(udo_scheme)
        if isinstance(udo_scheme, Reference):
            _udo_scheme_reference_validations(udo_scheme)
    for transformation in transformation_scheme.items:
        if not isinstance(transformation, Transformation):
            raise Invalid(
                "TransformationScheme must contain Transformation items"
            )
        _transformation_validations(transformation)



# mypy: disable-error-code="union-attr"
"""Module for writing SDMX-ML 2.1 Generic data messages."""

from typing import Any, Dict, List, Optional, Sequence, Tuple

import pandas as pd

from pysdmx.io.format import Format
from pysdmx.io.pd import PandasDataset
from pysdmx.io.xml.sdmx21.writer.__write_aux import (
    ABBR_GEN,
    ABBR_MSG,
    ALL_DIM,
    __escape_xml,
    __write_header,
    create_namespaces,
    get_end_message,
    get_structure,
)
from pysdmx.io.xml.sdmx21.writer.__write_data_aux import (
    check_content_dataset,
    check_dimension_at_observation,
    get_codes,
    writing_validation,
)
from pysdmx.io.xml.sdmx21.writer.config import CHUNKSIZE
from pysdmx.model.message import Header
from pysdmx.util import parse_short_urn


def __value(id: str, value: str) -> str:
    """Write a value tag."""
    return f"<{ABBR_GEN}:Value id={id!r} value={__escape_xml(str(value))!r}/>"


def __generate_obs_structure(
    dataset: PandasDataset,
) -> Tuple[List[str], str, List[str]]:
    """Generate the structure of the observations.

    First element is a list of
    the codes for Key, second element is the code for the value, and the third
    element is a list of the codes for the attributes.

    Args:
        dataset: PandasDataset. The dataset to generate the structure from.

    Returns:
        The structure of the observations
    """
    obs_structure: Tuple[List[str], str, List[str]] = (
        [],
        dataset.structure.components.measures[0].id,
        [],
    )
    for dim in dataset.structure.components.dimensions:
        obs_structure[0].append(dim.id)

    cols = dataset.data.columns

    for att in dataset.structure.components.attributes:
        if att.attachment_level == "O" and att.id in cols:
            obs_structure[2].append(att.id)
        elif (
            att.attachment_level is not None
            and att.attachment_level != "D"
            and att.id in cols
        ):
            obs_structure[0].append(att.id)

    return obs_structure


def __memory_optimization_writing(
    dataset: PandasDataset,
    obs_structure: Tuple[List[str], str, List[str]],
    prettyprint: bool,
) -> str:
    """Memory optimization for writing data."""
    outfile = ""
    length_ = len(dataset.data)
    if len(dataset.data) > CHUNKSIZE:
        previous = 0
        next_ = CHUNKSIZE
        while previous <= length_:
            # Sliding a window for efficient access to the data
            # and avoid memory issues
            outfile += __obs_processing(
                dataset.data.iloc[previous:next_], obs_structure, prettyprint
            )
            previous = next_
            next_ += CHUNKSIZE

            if next_ >= length_:
                outfile += __obs_processing(
                    dataset.data.iloc[previous:], obs_structure, prettyprint
                )
                previous = next_
    else:
        outfile += __obs_processing(dataset.data, obs_structure, prettyprint)

    return outfile


def __write_data_generic(
    datasets: Dict[str, PandasDataset],
    dim_mapping: Dict[str, str],
    prettyprint: bool = True,
) -> str:
    """Write data to SDMX-ML 2.1 Generic format.

    Args:
        datasets: dict. Datasets to be written.
        dim_mapping: dict. URN-DimensionAtObservation mapping.
        prettyprint: bool. Prettyprint or not.

    Returns:
        The data in SDMX-ML 2.1 Generic format, as string.
    """
    outfile = ""

    for short_urn, dataset in datasets.items():
        writing_validation(dataset)
        dataset.data = dataset.data.fillna("").astype(str)
        outfile += __write_data_single_dataset(
            dataset=dataset,
            prettyprint=prettyprint,
            dim=dim_mapping[short_urn],
        )

    return outfile


def __write_data_single_dataset(
    dataset: PandasDataset,
    prettyprint: bool = True,
    dim: str = ALL_DIM,
) -> str:
    """Write data to SDMX-ML 2.1 Generic format.

    Args:
        dataset: PandasDataset. Dataset to be written.
        prettyprint: bool. Prettyprint or not.
        dim: str. Dimension to be written.

    Returns:
        The data in SDMX-ML 2.1 Generic format, as string.
    """

    def __remove_optional_attributes_empty_data(str_to_check: str) -> str:
        """This function removes data when optional attributes are found."""
        for att in dataset.structure.components.attributes:
            if not att.required:
                to_replace = f'<{ABBR_GEN}:Value id={att.id!r} value=""/>'
                to_replace = f"{child3}{to_replace}{nl}"
                str_to_check = str_to_check.replace(to_replace, "")
        return str_to_check

    outfile = ""
    structure_urn = get_structure(dataset)
    id_structure = parse_short_urn(structure_urn).id

    nl = "\n" if prettyprint else ""
    child1 = "\t" if prettyprint else ""
    child2 = "\t\t" if prettyprint else ""
    child3 = "\t\t\t" if prettyprint else ""

    # Datasets
    outfile += (
        f"{nl}{child1}<{ABBR_MSG}:DataSet "
        f"structureRef={id_structure!r} "
        f"action={dataset.action.value!r}>{nl}"
    )
    data = ""
    # Write attached attributes
    attached_attributes_str = []
    for k, v in dataset.attributes.items():
        attached_attributes_str.append(__value(k, v))

    if len(attached_attributes_str) > 0:
        data += f"{child2}<{ABBR_GEN}:Attributes>{nl}"
        for att in attached_attributes_str:
            data += f"{child3}{att}{nl}"
        data += f"{child2}</{ABBR_GEN}:Attributes>{nl}"

    if dim == ALL_DIM:
        obs_structure = __generate_obs_structure(dataset)
        data += __memory_optimization_writing(
            dataset=dataset,
            obs_structure=obs_structure,
            prettyprint=prettyprint,
        )
    else:
        series_codes, obs_codes = get_codes(
            dimension_code=dim,
            structure=dataset.structure,  # type: ignore[arg-type]
            data=dataset.data,
        )
        att_codes = [att.id for att in dataset.structure.components.attributes]
        series_att_codes = [x for x in series_codes if x in att_codes]
        obs_att_codes = [x for x in obs_codes if x in att_codes]

        series_codes = [x for x in series_codes if x not in series_att_codes]
        obs_codes = [x for x in obs_codes if x not in obs_att_codes]

        data += __series_processing(
            data=dataset.data,
            series_codes=series_codes,
            series_att_codes=series_att_codes,
            obs_codes=obs_codes,
            obs_att_codes=obs_att_codes,
            prettyprint=prettyprint,
        )

    # Remove optional attributes empty data
    data = __remove_optional_attributes_empty_data(data)

    # Add to outfile
    outfile += data

    outfile += f"{child1}</{ABBR_MSG}:DataSet>"

    return outfile.replace("'", '"')


def __obs_processing(
    data: pd.DataFrame,
    obs_structure: Tuple[List[str], str, List[str]],
    prettyprint: bool = True,
) -> str:
    def __format_obs_str(element: Dict[str, Any]) -> str:
        child2 = "\t\t" if prettyprint else ""
        child3 = "\t\t\t" if prettyprint else ""
        child4 = "\t\t\t\t" if prettyprint else ""
        nl = "\n" if prettyprint else ""

        out = f"{child2}<{ABBR_GEN}:Obs>{nl}"
        # Obs Key writing
        out += f"{child3}<{ABBR_GEN}:ObsKey>{nl}"
        for k, v in element.items():
            if k in obs_structure[0]:
                out += f"{child4}{__value(k, v)}{nl}"
        out += f"{child3}</{ABBR_GEN}:ObsKey>{nl}"

        # Obs Value writing
        out += (
            f"{child3}<{ABBR_GEN}:ObsValue "
            f"value={str(element[obs_structure[1]])!r}/>{nl}"
        )

        if len(obs_structure[2]) > 0:
            # Obs Attributes writing
            out += f"{child3}<{ABBR_GEN}:Attributes>{nl}"
            for k, v in element.items():
                if k in obs_structure[2]:
                    out += f"{child4}{__value(k, v)}{nl}"
            out += f"{child3}</{ABBR_GEN}:Attributes>{nl}"

        out += f"{child2}</{ABBR_GEN}:Obs>{nl}"

        return out

    parser = lambda x: __format_obs_str(x)  # noqa: E731

    iterator = map(parser, data.to_dict(orient="records"))

    return "".join(iterator)


def __series_processing(
    data: pd.DataFrame,
    series_codes: List[str],
    series_att_codes: List[str],
    obs_codes: List[str],
    obs_att_codes: List[str],
    prettyprint: bool = True,
) -> str:
    def __generate_series_str() -> str:
        out_list: List[str] = []
        data.groupby(by=series_codes + series_att_codes).apply(
            lambda x: __format_dict_ser(out_list, x)
        )

        return "".join(out_list)

    def __format_dict_ser(
        output_list: List[str],
        group_data: Any,
    ) -> Any:
        obs_data = group_data[obs_codes + obs_att_codes].copy()
        data_dict["Series"][0]["Obs"] = obs_data.to_dict(orient="records")
        data_dict["Series"][0].update(
            {
                k: v
                for k, v in group_data[series_att_codes].iloc[0].items()
                if k in series_att_codes
            }
        )
        output_list.append(
            __format_ser_str(
                data_info=data_dict["Series"][0],
                series_codes=series_codes,
                series_att_codes=series_att_codes,
                obs_codes=obs_codes,
                obs_att_codes=obs_att_codes,
                prettyprint=prettyprint,
            )
        )
        del data_dict["Series"][0]

    # Getting each datapoint from data and creating dict
    data = data.sort_values(series_codes, axis=0)
    data_dict = {
        "Series": data[series_codes]
        .drop_duplicates()
        .reset_index(drop=True)
        .to_dict(orient="records")
    }

    out = __generate_series_str()

    return out


def __format_ser_str(
    data_info: Dict[Any, Any],
    series_codes: List[str],
    series_att_codes: List[str],
    obs_codes: List[str],
    obs_att_codes: List[str],
    prettyprint: bool,
) -> str:
    child2 = "\t\t" if prettyprint else ""
    child3 = "\t\t\t" if prettyprint else ""
    child4 = "\t\t\t\t" if prettyprint else ""
    child5 = "\t\t\t\t\t" if prettyprint else ""
    nl = "\n" if prettyprint else ""

    out_element = f"{child2}<{ABBR_GEN}:Series>{nl}"

    # Series Key writing
    out_element += f"{child3}<{ABBR_GEN}:SeriesKey>{nl}"
    for k, v in data_info.items():
        if k in series_codes:
            out_element += f"{child4}{__value(k, v)}{nl}"
    out_element += f"{child3}</{ABBR_GEN}:SeriesKey>{nl}"

    # Series Attributes writing
    if len(series_att_codes) > 0:
        out_element += f"{child3}<{ABBR_GEN}:Attributes>{nl}"
        for k, v in data_info.items():
            if k in series_att_codes:
                out_element += f"{child4}{__value(k, v)}{nl}"
        out_element += f"{child3}</{ABBR_GEN}:Attributes>{nl}"

    # Obs writing
    for obs in data_info["Obs"]:
        out_element += f"{child3}<{ABBR_GEN}:Obs>{nl}"

        # Obs Dimension writing
        out_element += (
            f"{child4}<{ABBR_GEN}:ObsDimension "
            f"value={str(obs[obs_codes[0]])!r}/>{nl}"
        )
        # Obs Value writing
        out_element += (
            f"{child4}<{ABBR_GEN}:ObsValue value={str(obs_codes[1])!r}/>{nl}"
        )

        # Obs Attributes writing
        if len(obs_att_codes) > 0:
            out_element += f"{child4}<{ABBR_GEN}:Attributes>{nl}"
            for k, v in obs.items():
                if k in obs_att_codes:
                    out_element += f"{child5}{__value(k, v)}{nl}"
            out_element += f"{child4}</{ABBR_GEN}:Attributes>{nl}"
        out_element += f"{child3}</{ABBR_GEN}:Obs>{nl}"

    out_element += f"{child2}</{ABBR_GEN}:Series>{nl}"

    return out_element


def write(
    datasets: Sequence[PandasDataset],
    output_path: str = "",
    prettyprint: bool = True,
    header: Optional[Header] = None,
    dimension_at_observation: Optional[Dict[str, str]] = None,
) -> Optional[str]:
    """Write data to SDMX-ML 2.1 Generic format.

    Args:
        datasets: The datasets to be written.
        output_path: The path to save the file.
        prettyprint: Prettyprint or not.
        header: The header to be used (generated if None).
        dimension_at_observation:
          The mapping between the dataset and the dimension at observation.

    Returns:
        The XML string if path is empty, None otherwise.
    """
    type_ = Format.DATA_SDMX_ML_2_1_GEN

    # Checking if we have datasets,
    # we need to ensure we can write them correctly
    check_content_dataset(datasets)
    content = {dataset.short_urn: dataset for dataset in datasets}

    if header is None:
        header = Header()

    # Checking the dimension at observation mapping
    dim_mapping = check_dimension_at_observation(
        datasets=content, dimension_at_observation=dimension_at_observation
    )
    header.structure = dim_mapping
    # Generating the initial tag with namespaces
    outfile = create_namespaces(type_, prettyprint=prettyprint)
    # Generating the header
    outfile += __write_header(header, prettyprint, data_message=True)
    # Writing the content
    outfile += __write_data_generic(content, dim_mapping, prettyprint)

    outfile += get_end_message(type_, prettyprint)

    if output_path == "":
        return outfile

    with open(output_path, "w", encoding="UTF-8", errors="replace") as f:
        f.write(outfile)

    return None



"""Validates an SDMX-ML 2.1 XML file against the XSD schema."""

from io import BytesIO

from lxml import etree
from sdmxschemas import SDMX_ML_21_MESSAGE_PATH as SCHEMA_PATH

from pysdmx.errors import Invalid
from pysdmx.io.xml.__allowed_lxml_errors import ALLOWED_ERRORS_CONTENT


def validate_doc(input_str: str) -> None:
    """Validates the SDMX-ML data against the XSD schema for SDMX-ML 2.1.

    Args:
        input_str: The SDMX-ML data to validate.

    Raises:
        Invalid: If the SDMX-ML data does not validate against the schema.
    """
    parser = etree.ETCompatXMLParser()
    xmlschema_doc = etree.parse(SCHEMA_PATH)
    xmlschema = etree.XMLSchema(xmlschema_doc)

    bytes_infile = BytesIO(bytes(input_str, "UTF_8"))

    doc = etree.parse(bytes_infile, parser=parser)
    if not xmlschema.validate(doc):
        log_errors = list(xmlschema.error_log)  # type: ignore[call-overload]
        unhandled_errors = []
        for e in log_errors:
            unhandled_errors.append(e.message)
        severe_errors = unhandled_errors.copy()
        for e in unhandled_errors:
            for allowed_error in ALLOWED_ERRORS_CONTENT:
                if allowed_error in e:
                    severe_errors.remove(e)

        if len(severe_errors) > 0:
            raise Invalid("Validation Error", ";\n".join(severe_errors))



"""API to be implemented by connectors."""

from datetime import datetime
from typing import (
    Any,
    Generator,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Union,
    runtime_checkable,
)

from pysdmx.api.dc.query import (
    BooleanFilter,
    DateTimeFilter,
    MultiFilter,
    NotFilter,
    NumberFilter,
    SortBy,
    TextFilter,
)
from pysdmx.model import (
    Dataflow,
    DataflowInfo,
    DataflowRef,
    Organisation,
    SeriesInfo,
)


@runtime_checkable
class Connector(Protocol):
    """A Connector provides access to a data source.

    Connectors support both data discovery and data retrieval.

    Attributes:
        owner: The organisation providing the Connector.
            It is recommended to provide contact details, in case users
            of the Connector have questions about it.
    """

    owner: Organisation

    def dataflows(
        self,
        filter_query: Optional[str] = None,
        provider: Optional[str] = None,
    ) -> Union[Iterable[Dataflow], Iterable[DataflowRef]]:
        """Get the dataflows provided by the Connector.

        Args:
            filter_query: A search term. If set, any dataflow containing the
                term in its id, name or description will be returned.
            provider: A data provider. If set, any dataflow for which data are
                being provided by the supplied provider will be returned.

        Returns: Iterable[DataflowRef]: A collection of dataflow references.

                The references contain core information about the
                dataflows provided by the Connector. Most importantly,
                they contain the identifying information required to
                retrieve more information about a specific dataflow
                (see dataflow).

                It is expected that this method, if implemented,
                will return at least one DataflowRef object.
        """

    def providers(
        self,
        filter_query: Optional[str] = None,
    ) -> Iterable[Organisation]:
        """Get the providers of which data are available in the Connector.

        Args:
            filter_query: A search term. If set, any provider containing the
                term in its id, name or description will be returned.

        Returns: Iterable[Organisation]: A collection of data providers.
        """

    def dataflow(
        self,
        dataflow: Union[str, DataflowRef, Dataflow],
        metrics: bool = False,
    ) -> DataflowInfo:
        """Get information about a dataflow.

        Args:
            dataflow: Either a string representing the dataflow ID or
                a DataflowRef, as returned by calling the dataflows
                function.
            metrics: Metrics contain useful information such as
                the number of observations, series, when the
                dataflow was last updated, etc. However, it may be
                expensive to fetch this information. In case metrics
                are not required, you may set metrics to False, to
                prevent them from being computed.

        Returns: DataflowInfo: Information about the requested dataflow.

            The information includes:

            - Some basic metadata about the dataflow (such as its ID and name).
            - Some useful metrics such as the number of observations.
            - The expected structure of data (i.e. the data schema), including
              the expected columns, their types, etc.
        """

    def series(
        self,
        dataflow: Union[str, DataflowRef, Dataflow, DataflowInfo],
        provider: Optional[Union[str, Organisation]] = None,
        filters: Optional[
            Union[
                BooleanFilter,
                DateTimeFilter,
                MultiFilter,
                NotFilter,
                NumberFilter,
                str,
                TextFilter,
            ]
        ] = None,
        discontinued: bool = False,
        updated_after: Optional[datetime] = None,
    ) -> Generator[SeriesInfo, None, None]:
        """Get the series available in the dataflow.

        Args:
            dataflow: Either a string representing the dataflow ID or
                a DataflowRef, as returned by calling the dataflows
                function, or a DataflowInfo, as returned by calling the
                dataflow function.
            provider: The organisation providing the data to be returned.
            filters: A dictionary, where the keys are the IDs of the
                components used for filtering and the values
                represent the filters to be applied.
            discontinued: Whether discontinued (i.e. old)
                series should be returned. Defaults to `False`.
            updated_after: Retrieve the series updated after
                the supplied timestamp.

        Returns: A generator, to iterate over the collection matching series
            The information includes:

            - Some basic metadata about the series (such as an ID or name).
            - Some useful metrics such as the number of observations, the last
              time it was updated, etc.
        """

    def data(
        self,
        dataflow: Union[str, DataflowRef, Dataflow, DataflowInfo],
        provider: Optional[Union[str, Organisation]] = None,
        series: Optional[Iterable[str]] = None,
        filters: Optional[
            Union[
                BooleanFilter,
                DateTimeFilter,
                MultiFilter,
                NotFilter,
                NumberFilter,
                str,
                TextFilter,
            ]
        ] = None,
        columns: Optional[Iterable[str]] = None,
        sort: Optional[List[SortBy]] = None,
        offset: int = 0,
        limit: Optional[int] = None,
        history: bool = False,
        updated_after: Optional[datetime] = None,
        format: Literal["dict", "pandas"] = "dict",
    ) -> Any:
        """Get data for the selected dataflow.

        Args:
            dataflow: Either a string representing the dataflow ID or
                a DataflowRef, as returned by calling the dataflows
                function, or a DataflowInfo, as returned by calling the
                dataflow function.
            provider: The organisation providing the data to be returned.
            series: One or more strings identifying the series
                for which data should be returned (e.g. D.USD.EUR.SP00).
                wildcards can be used too (e.g. D.*.EUR.SP00).
            filters: A dictionary, where the keys are the IDs of the
                components used for filtering and the values
                represent the filters to be applied.
            columns: The fields to be returned.
            sort: The desired sorting order of the results.
            offset: From which row to start fetching. This is useful to
                paginate results. Default to 0.
            limit: The number of rows to be returned.
            history: Whether to retrieve previous "versions" of the data or
                only the latest one. This is useful to see how the data may
                have evolved or have been corrected over time.
            updated_after: Retrieve the series updated after
                the supplied timestamp.
            format: The desired format for the data. By default, a generator
                of dict objects will be returned, but there are other options
                such as pandas data frames.

        Returns: A generator, to iterate over the matching data, if format is
            set to dict (the default), or a pandas data frame.
        """


__all__ = ["Connector"]



"""Lists of supported formats."""

from enum import Enum

_BASE = "application/vnd.sdmx."


class Format(Enum):
    """The list of SDMX formats since version 2.1."""

    DATA_SDMX_CSV_1_0_0 = f"{_BASE}data+csv;version=1.0.0"
    DATA_SDMX_CSV_2_0_0 = f"{_BASE}data+csv;version=2.0.0"
    DATA_SDMX_JSON_1_0_0 = f"{_BASE}data+json;version=1.0.0"
    DATA_SDMX_JSON_2_0_0 = f"{_BASE}data+json;version=2.0.0"
    DATA_SDMX_ML_2_1_GEN = f"{_BASE}genericdata+xml;version=2.1"
    DATA_SDMX_ML_2_1_STR = f"{_BASE}structurespecificdata+xml;version=2.1"
    DATA_SDMX_ML_2_1_GENTS = f"{_BASE}generictimeseriesdata+xml;version=2.1"
    DATA_SDMX_ML_2_1_STRTS = (
        f"{_BASE}structurespecifictimeseriesdata+xml;version=2.1"
    )
    DATA_SDMX_ML_3_0 = f"{_BASE}data+xml;version=3.0.0"
    REFMETA_SDMX_CSV_2_0_0 = f"{_BASE}metadata+csv;version=2.0.0"
    REFMETA_SDMX_JSON_2_0_0 = f"{_BASE}metadata+json;version=2.0.0"
    REFMETA_SDMX_ML_3_0 = f"{_BASE}metadata+xml;version=3.0.0"
    REGISTRY_SDMX_ML_2_1 = f"{_BASE}registry+xml;version=2.1"
    REGISTRY_SDMX_ML_3_0 = f"{_BASE}registry+xml;version=3.0"
    SCHEMA_SDMX_JSON_1_0_0 = f"{_BASE}schema+json;version=1.0.0"
    SCHEMA_SDMX_JSON_2_0_0 = f"{_BASE}schema+json;version=2.0.0"
    SCHEMA_SDMX_ML_2_1 = f"{_BASE}schema+xml;version=2.1"
    SCHEMA_SDMX_ML_3_0 = f"{_BASE}schema+xml;version=3.0.0"
    STRUCTURE_SDMX_JSON_1_0_0 = f"{_BASE}structure+json;version=1.0.0"
    STRUCTURE_SDMX_JSON_2_0_0 = f"{_BASE}structure+json;version=2.0.0"
    STRUCTURE_SDMX_ML_2_1 = f"{_BASE}structure+xml;version=2.1"
    STRUCTURE_SDMX_ML_3_0 = f"{_BASE}structure+xml;version=3.0.0"
    ERROR_SDMX_ML_2_1 = "application/error.xml"
    FUSION_JSON = "application/vnd.fusion.json"


class AvailabilityFormat(Enum):
    """The SDMX Availability formats."""

    SDMX_JSON_1_0_0 = Format.STRUCTURE_SDMX_JSON_1_0_0.value
    SDMX_JSON_2_0_0 = Format.STRUCTURE_SDMX_JSON_2_0_0.value
    SDMX_ML_2_1 = Format.STRUCTURE_SDMX_ML_2_1.value
    SDMX_ML_3_0 = Format.STRUCTURE_SDMX_ML_3_0.value


class DataFormat(Enum):
    """The SDMX Data formats."""

    SDMX_CSV_1_0_0 = Format.DATA_SDMX_CSV_1_0_0.value
    SDMX_CSV_2_0_0 = Format.DATA_SDMX_CSV_2_0_0.value
    SDMX_JSON_1_0_0 = Format.DATA_SDMX_JSON_1_0_0.value
    SDMX_JSON_2_0_0 = Format.DATA_SDMX_JSON_2_0_0.value
    SDMX_ML_2_1_GEN = Format.DATA_SDMX_ML_2_1_GEN.value
    SDMX_ML_2_1_GENTS = Format.DATA_SDMX_ML_2_1_GENTS.value
    SDMX_ML_2_1_STR = Format.DATA_SDMX_ML_2_1_STR.value
    SDMX_ML_2_1_STRTS = Format.DATA_SDMX_ML_2_1_STRTS.value
    SDMX_ML_3_0 = Format.DATA_SDMX_ML_3_0.value


class RefMetaFormat(Enum):
    """The SDMX Reference Metadata formats."""

    SDMX_CSV_2_0_0 = Format.REFMETA_SDMX_CSV_2_0_0.value
    SDMX_JSON_2_0_0 = Format.REFMETA_SDMX_JSON_2_0_0.value
    SDMX_ML_3_0 = Format.REFMETA_SDMX_ML_3_0.value
    FUSION_JSON = Format.FUSION_JSON.value


class SchemaFormat(Enum):
    """The SDMX Schema formats."""

    SDMX_JSON_1_0_0_SCHEMA = Format.SCHEMA_SDMX_JSON_1_0_0.value
    SDMX_JSON_2_0_0_SCHEMA = Format.SCHEMA_SDMX_JSON_2_0_0.value
    SDMX_JSON_1_0_0_STRUCTURE = Format.STRUCTURE_SDMX_JSON_1_0_0.value
    SDMX_JSON_2_0_0_STRUCTURE = Format.STRUCTURE_SDMX_JSON_2_0_0.value
    SDMX_ML_2_1_SCHEMA = Format.SCHEMA_SDMX_ML_2_1.value
    SDMX_ML_3_0_SCHEMA = Format.SCHEMA_SDMX_ML_3_0.value
    SDMX_ML_2_1_STRUCTURE = Format.STRUCTURE_SDMX_ML_2_1.value
    SDMX_ML_3_0_STRUCTURE = Format.STRUCTURE_SDMX_ML_3_0.value
    FUSION_JSON = Format.FUSION_JSON.value


class StructureFormat(Enum):
    """The SDMX Structure formats."""

    SDMX_JSON_1_0_0 = Format.STRUCTURE_SDMX_JSON_1_0_0.value
    SDMX_JSON_2_0_0 = Format.STRUCTURE_SDMX_JSON_2_0_0.value
    SDMX_ML_2_1 = Format.STRUCTURE_SDMX_ML_2_1.value
    SDMX_ML_3_0 = Format.STRUCTURE_SDMX_ML_3_0.value
    FUSION_JSON = Format.FUSION_JSON.value


class RegistryFormat(Enum):
    """The SDMX Registry formats."""

    SDMX_ML_2_1 = Format.REGISTRY_SDMX_ML_2_1.value
    SDMX_ML_3_0 = Format.REGISTRY_SDMX_ML_3_0.value
    FUSION_JSON = Format.FUSION_JSON.value
